import re

from djblets.conditions.choices import BaseConditionChoice
from djblets.conditions.errors import ConditionOperatorNotFoundError
from djblets.conditions.operators import (AnyOperator,
                                          BaseConditionOperator,
                                          ConditionOperators,
                                          ContainsAnyOperator,
                                          ContainsOperator,
                                          DoesNotContainAnyOperator,
                                          DoesNotContainOperator,
                                          DoesNotMatchRegexOperator,
                                          EndsWithOperator,
                                          GreaterThanOperator,
                                          IsNotOneOfOperator,
                                          IsNotOperator,
                                          IsOneOfOperator,
                                          IsOperator,
                                          LessThanOperator,
                                          MatchesRegexOperator,
                                          StartsWithOperator,
                                          UnsetOperator)
from djblets.conditions.values import (BaseConditionValueField,
                                       ConditionValueIntegerField)
from djblets.testing.testcases import TestCase


class BaseConditionOperatorTests(TestCase):
    """Unit tests for djblets.conditions.operators.BaseConditionOperator."""

    def test_value_field_with_default(self):
        """Testing BaseConditionOperator.value_field with default choice field
        """
        class MyOperator(BaseConditionOperator):
            operator_id = 'my-op'

        class MyChoice(BaseConditionChoice):
            operators = ConditionOperators([
                MyOperator,
            ])

            default_value_field = BaseConditionValueField()

        choice = MyChoice()
        op = choice.get_operator('my-op')
        self.assertEqual(op.value_field, choice.default_value_field)
        self.assertFalse(op.has_custom_value_field)

    def test_value_field_with_default_as_function(self):
        """Testing BaseConditionOperator.value_field with default choice field
        as a function
        """
        class MyOperator(BaseConditionOperator):
            operator_id = 'my-op'

        class MyChoice(BaseConditionChoice):
            operators = ConditionOperators([
                MyOperator,
            ])

            def default_value_field(self, **kwargs):
                return _my_value_field

        _my_value_field = BaseConditionValueField()

        choice = MyChoice()
        op = choice.get_operator('my-op')
        self.assertEqual(op.value_field, _my_value_field)
        self.assertFalse(op.has_custom_value_field)

    def test_value_field_with_custom(self):
        """Testing BaseConditionOperator.value_field with custom choice field
        """
        class MyOperator(BaseConditionOperator):
            operator_id = 'my-op'

            value_field = BaseConditionValueField()

        class MyChoice(BaseConditionChoice):
            operators = ConditionOperators([
                MyOperator,
            ])

            default_value_field = BaseConditionValueField()

        choice = MyChoice()
        op = choice.get_operator('my-op')
        self.assertNotEqual(op.value_field, choice.default_value_field)
        self.assertTrue(op.has_custom_value_field)

    def test_with_overrides(self):
        """Testing BaseConditionOperator.with_overrides"""
        class MyOperator(BaseConditionOperator):
            operator_id = 'my-op'
            name = 'My Op'

            value_field = BaseConditionValueField()

        CustomOperator = MyOperator.with_overrides(
            name='Custom Op',
            value_field=ConditionValueIntegerField())

        self.assertEqual(CustomOperator.__name__, 'CustomMyOperator')
        self.assertEqual(CustomOperator.name, 'Custom Op')
        self.assertIs(CustomOperator.value_field.__class__,
                      ConditionValueIntegerField)


class ConditionOperatorsTests(TestCase):
    """Unit tests for djblets.conditions.operators.ConditionOperators."""

    def test_init_with_class_operators(self):
        """Testing ConditionOperators initialization with class-defined operators
        """
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-operator-1'

        class MyOperators(ConditionOperators):
            operator_classes = [
                MyOperator1,
            ]

        operators = MyOperators()
        self.assertEqual(list(operators), [MyOperator1])

    def test_init_with_caller_operators(self):
        """Testing ConditionOperators initialization with caller-defined operators
        """
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-operator-1'

        class MyOperator2(BaseConditionOperator):
            operator_id = 'my-operator-2'

        class MyOperators(ConditionOperators):
            operator_classes = [
                MyOperator1,
            ]

        operators = MyOperators([MyOperator2])
        self.assertEqual(list(operators), [MyOperator2])

    def test_get_operator(self):
        """Testing ConditionOperators.get_operator"""
        class MyOperator1(BaseConditionOperator):
            operator_id = 'my-operator-1'

        class MyChoice(BaseConditionChoice):
            operators = ConditionOperators([MyOperator1])

        choice = MyChoice()
        self.assertEqual(
            choice.operators.get_operator('my-operator-1', choice).__class__,
            MyOperator1)

    def test_get_operator_with_invalid_id(self):
        """Testing ConditionOperators.get_operator with invalid ID"""
        operators = ConditionOperators()

        with self.assertRaises(ConditionOperatorNotFoundError):
            operators.get_operator('invalid', None)


class StandardOperatorTests(TestCase):
    """Unit tests for standard condition operators."""

    def test_is_one_of_op_with_match(self):
        """Testing IsOneOfOperator with match"""
        self.assertTrue(self._check_match(IsOneOfOperator, 'a',
                                          ['a', 'b', 'c']))
        self.assertTrue(self._check_match(IsOneOfOperator, 1, [0, 1, 2]))

    def test_is_one_of_op_without_match(self):
        """Testing IsOneOfOperator without match"""
        self.assertFalse(self._check_match(IsOneOfOperator, 'd',
                                           ['a', 'b', 'c']))
        self.assertFalse(self._check_match(IsOneOfOperator, 4, [0, 1, 2]))

    def test_is_not_one_of_op_with_match(self):
        """Testing IsNotOneOfOperator with match"""
        self.assertTrue(self._check_match(IsNotOneOfOperator, 'z',
                                          ['a', 'b', 'c']))
        self.assertTrue(self._check_match(IsNotOneOfOperator, 9, [0, 1, 2]))

    def test_is_not_one_of_op_without_match(self):
        """Testing IsNotOneOfOperator without match"""
        self.assertFalse(self._check_match(IsNotOneOfOperator, 'a',
                                           ['a', 'b', 'c']))
        self.assertFalse(self._check_match(IsNotOneOfOperator, 0, [0, 1, 2]))

    def test_any_op_with_match(self):
        """Testing AnyOperator with match"""
        self.assertTrue(self._check_match(AnyOperator, 'foo'))
        self.assertTrue(self._check_match(AnyOperator, 0))
        self.assertTrue(self._check_match(AnyOperator, False))

    def test_any_op_without_match(self):
        """Testing AnyOperator without match"""
        self.assertFalse(self._check_match(AnyOperator, ''))
        self.assertFalse(self._check_match(AnyOperator, None))
        self.assertFalse(self._check_match(AnyOperator, {}))
        self.assertFalse(self._check_match(AnyOperator, []))

    def test_unset_op_with_match(self):
        """Testing UnsetOperator with match"""
        self.assertTrue(self._check_match(UnsetOperator, ''))
        self.assertTrue(self._check_match(UnsetOperator, None))
        self.assertTrue(self._check_match(UnsetOperator, {}))
        self.assertTrue(self._check_match(UnsetOperator, []))

    def test_unset_op_without_match(self):
        """Testing UnsetOperator without match"""
        self.assertFalse(self._check_match(UnsetOperator, 'foo'))
        self.assertFalse(self._check_match(UnsetOperator, 0))
        self.assertFalse(self._check_match(UnsetOperator, False))

    def test_is_op_with_match(self):
        """Testing IsOperator with match"""
        self.assertTrue(self._check_match(IsOperator, 1, 1))
        self.assertTrue(self._check_match(IsOperator, 'test', 'test'))
        self.assertTrue(self._check_match(IsOperator, True, True))
        self.assertTrue(self._check_match(IsOperator, ['a', 'b'], ['a', 'b']))

    def test_is_op_without_match(self):
        """Testing IsOperator without match"""
        self.assertFalse(self._check_match(IsOperator, 0, 1))
        self.assertFalse(self._check_match(IsOperator, 'hi', 'test'))
        self.assertFalse(self._check_match(IsOperator, True, False))
        self.assertFalse(self._check_match(IsOperator, [1, 2], ['a', 'b']))

    def test_is_not_op_with_match(self):
        """Testing IsNotOperator with match"""
        self.assertTrue(self._check_match(IsNotOperator, 0, 1))
        self.assertTrue(self._check_match(IsNotOperator, 'hi', 'test'))
        self.assertTrue(self._check_match(IsNotOperator, True, False))
        self.assertTrue(self._check_match(IsNotOperator, [1, 2], ['a', 'b']))

    def test_is_not_op_without_match(self):
        """Testing IsNotOperator without match"""
        self.assertFalse(self._check_match(IsNotOperator, 1, 1))
        self.assertFalse(self._check_match(IsNotOperator, 'test', 'test'))
        self.assertFalse(self._check_match(IsNotOperator, True, True))
        self.assertFalse(self._check_match(IsNotOperator,
                                           ['a', 'b'], ['a', 'b']))

    def test_contains_op_with_match(self):
        """Testing ContainsOperator with match"""
        self.assertTrue(self._check_match(ContainsOperator,
                                          'hello world', 'world'))
        self.assertTrue(self._check_match(ContainsOperator, [1, 2, 3], 1))

    def test_contains_op_without_match(self):
        """Testing ContainsOperator without match"""
        self.assertFalse(self._check_match(ContainsOperator,
                                           'hello world', 'hey'))
        self.assertFalse(self._check_match(ContainsOperator, [1, 2, 3], 4))

    def test_does_not_contain_op_with_match(self):
        """Testing DoesNotContainOperator with match"""
        self.assertTrue(self._check_match(DoesNotContainOperator,
                                          'hello world', 'hey'))
        self.assertTrue(self._check_match(DoesNotContainOperator,
                                          [1, 2, 3], 4))

    def test_does_not_contain_op_without_match(self):
        """Testing DoesNotContainOperator without match"""
        self.assertFalse(self._check_match(DoesNotContainOperator,
                                           'hello world', 'hello'))
        self.assertFalse(self._check_match(DoesNotContainOperator,
                                           [1, 2, 3], 1))

    def test_contains_any_op_with_match(self):
        """Testing ContainsAnyOperator with match"""
        self.assertTrue(self._check_match(ContainsAnyOperator,
                                          ['abc', 'def', 'xyz'],
                                          ['def']))
        self.assertTrue(self._check_match(ContainsAnyOperator,
                                          ['abc', 'def', 'xyz'],
                                          ['def', 'XXX']))

    def test_contains_any_op_without_match(self):
        """Testing ContainsAnyOperator without match"""
        self.assertFalse(self._check_match(ContainsAnyOperator,
                                           ['abc', 'def', 'xyz'],
                                           ['XXX']))

    def test_does_not_contain_any_op_with_match(self):
        """Testing DoesNotContainAnyOperator with match"""
        self.assertTrue(self._check_match(DoesNotContainAnyOperator,
                                          ['abc', 'def', 'xyz'],
                                          ['XXX']))

    def test_does_not_contain_any_op_without_match(self):
        """Testing DoesNotContainAnyOperator without match"""
        self.assertFalse(self._check_match(DoesNotContainAnyOperator,
                                           ['abc', 'def', 'xyz'],
                                           ['def']))
        self.assertFalse(self._check_match(DoesNotContainAnyOperator,
                                           ['abc', 'def', 'xyz'],
                                           ['def', 'XXX']))

    def test_starts_with_op_with_match(self):
        """Testing StartsWithOperator with match"""
        self.assertTrue(self._check_match(StartsWithOperator,
                                          'hello world', 'he'))

    def test_starts_without_op_without_match(self):
        """Testing StartsWithOperator without match"""
        self.assertFalse(self._check_match(StartsWithOperator,
                                           'hello world', 'wo'))

    def test_ends_with_op_with_match(self):
        """Testing EndsWithOperator with match"""
        self.assertTrue(self._check_match(EndsWithOperator,
                                          'hello world', 'ld'))

    def test_ends_without_op_without_match(self):
        """Testing EndsWithOperator without match"""
        self.assertFalse(self._check_match(EndsWithOperator,
                                           'hello world', 'lo'))

    def test_greater_than_op_with_match(self):
        """Testing GreaterThanOperator with match"""
        self.assertTrue(self._check_match(GreaterThanOperator, 100, 20))

    def test_greater_than_op_without_match(self):
        """Testing GreaterThanOperator without match"""
        self.assertFalse(self._check_match(GreaterThanOperator, 20, 100))

    def test_less_than_op_with_match(self):
        """Testing LessThanOperator with match"""
        self.assertTrue(self._check_match(LessThanOperator, 20, 100))

    def test_less_than_op_without_match(self):
        """Testing LessThanOperator without match"""
        self.assertFalse(self._check_match(LessThanOperator, 100, 20))

    def test_matches_regex_op_with_match(self):
        """Testing MatchesRegexOperator with match"""
        self.assertTrue(self._check_match(
            MatchesRegexOperator,
            'abccd',
            re.compile('abc+de?')))

    def test_matches_regex_op_without_match(self):
        """Testing MatchesRegexOperator without match"""
        self.assertFalse(self._check_match(
            MatchesRegexOperator,
            'xyz',
            re.compile('abc+de?')))

    def test_does_not_match_regex_op_with_match(self):
        """Testing DoesNotMatchRegexOperator with match"""
        self.assertTrue(self._check_match(
            DoesNotMatchRegexOperator,
            'xyz',
            re.compile('abc+de?')))

    def test_does_not_match_regex_op_without_match(self):
        """Testing DoesNotMatchRegexOperator without match"""
        self.assertFalse(self._check_match(
            DoesNotMatchRegexOperator,
            'abccd',
            re.compile('abc+de?')))

    def _check_match(self, op_cls, match_value, condition_value=None):
        op = op_cls(None)
        return op.matches(match_value=match_value,
                          condition_value=condition_value)



"""Compatibility tags for django-pipeline <= 1.3.

django-pipeline 1.5 renamed the old ``compressed_css` and ``compressed_js``
template tags to ``stylesheet`` and ``javascript``. While this change makes a
lot of sense, it's possible that third-party users will still have templates
relying on the old names.
"""

from django import template
from pipeline.templatetags.pipeline import javascript, stylesheet


register = template.Library()


register.tag('compressed_css', stylesheet)
register.tag('compressed_js', javascript)



"""Unit tests for djblets.pipeline.compilers.less.LessCompiler."""

import os
import shutil
import tempfile

import kgb

from djblets.pipeline.compilers.less import LessCompiler
from djblets.testing.testcases import TestCase


class LessCompilerTests(kgb.SpyAgency, TestCase):
    """Unit tests for djblets.pipeline.compilers.less.LessCompiler."""

    @classmethod
    def setUpClass(cls):
        super(LessCompilerTests, cls).setUpClass()

        cls.compiler = LessCompiler(verbose=True,
                                    storage=None)

        tempdir = tempfile.mkdtemp(prefix='djblets-tests.')
        cls.tempdir = tempdir

        cls.main_less = os.path.join(tempdir, 'main.less')
        cls.dep1_less = os.path.join(tempdir, 'dep1.less')
        cls.dep2_less = os.path.join(tempdir, 'dep2.less')
        cls.dep3_less = os.path.join(tempdir, 'dep3.less')
        cls.main_css = os.path.join(tempdir, 'main.css')
        cls.dep3_css = os.path.join(tempdir, 'dep3.css')
        cls.dep_filenames = [
            cls.dep1_less,
            cls.dep2_less,
            cls.dep3_less,
        ]
        cls.main_filenames = [cls.main_less] + cls.dep_filenames
        cls.all_filenames = cls.main_filenames + [cls.main_css]

        with open(cls.main_less, 'w') as fp:
            fp.write('@import (reference) "dep1.less";\n')
            fp.write('@import (reference) "dep2.less";\n')

        with open(cls.dep1_less, 'w') as fp:
            fp.write('@abc: 123;\n')

        with open(cls.dep2_less, 'w') as fp:
            fp.write('@import (reference) "dep3.less";\n')

        with open(cls.dep3_less, 'w') as fp:
            fp.write('.foo() {}\n')

        with open(cls.main_css, 'w') as fp:
            fp.write('.foo() {}\n')

    @classmethod
    def tearDownClass(cls):
        super(LessCompilerTests, cls).tearDownClass()

        try:
            shutil.rmtree(cls.tempdir)
        except Exception:
            pass

        cls.compiler = None

    def tearDown(self):
        super(LessCompilerTests, self).tearDown()

        self.compiler.clear_caches()

        mtime = os.path.getmtime(self.main_css)

        for filename in self.main_filenames:
            os.utime(filename, (mtime, mtime))

    def test_is_outdated_with_up_to_date(self):
        """Testing LessCompiler.is_outdated with up-to-date files"""
        compiler = self.compiler
        main_less = self.main_less
        main_css = self.main_css

        getmtime_spy = self.spy_on(os.path.getmtime)
        get_import_files_spy = self.spy_on(compiler._get_import_files)

        self.assertFalse(compiler.is_outdated(main_less, main_css))
        self.assertEqual(
            compiler._import_files_cache,
            {
                main_less: set(self.dep_filenames),
            })

        self.assertEqual(set(compiler._mtime_cache),
                         set(self.all_filenames))

        self.assertSpyCallCount(getmtime_spy, 5)

        for filename in self.all_filenames:
            self.assertSpyCalledWith(getmtime_spy, filename)

        self.assertSpyCallCount(get_import_files_spy, 1)
        self.assertSpyCalledWith(get_import_files_spy, main_less)

        # Make sure everything stayed cached, since nothing was invalidated.
        getmtime_spy.reset_calls()
        get_import_files_spy.reset_calls()

        self.assertFalse(compiler.is_outdated(main_less, main_css))
        self.assertEqual(
            compiler._import_files_cache,
            {
                main_less: set(self.dep_filenames),
            })

        self.assertEqual(set(compiler._mtime_cache),
                         set(self.all_filenames))

        self.assertSpyNotCalled(getmtime_spy)
        self.assertSpyNotCalled(get_import_files_spy)

    def test_is_outdated_with_outdated_src_file(self):
        """Testing LessCompiler.is_outdated with outdated source file"""
        compiler = self.compiler
        main_less = self.main_less
        main_css = self.main_css

        getmtime_spy = self.spy_on(os.path.getmtime)
        get_import_files_spy = self.spy_on(compiler._get_import_files)

        self.assertFalse(compiler.is_outdated(main_less, main_css))
        self.assertEqual(
            compiler._import_files_cache,
            {
                main_less: set(self.dep_filenames),
            })

        self.assertEqual(set(compiler._mtime_cache),
                         set(self.all_filenames))

        self.assertSpyCallCount(getmtime_spy, 5)

        for filename in self.all_filenames:
            self.assertSpyCalledWith(getmtime_spy, filename)

        self.assertSpyCallCount(get_import_files_spy, 1)
        self.assertSpyCalledWith(get_import_files_spy, main_less)

        # Update the timestamp and try again.
        self._touch(main_less)

        getmtime_spy.reset_calls()
        get_import_files_spy.reset_calls()

        self.assertTrue(compiler.is_outdated(main_less, main_css))
        self.assertEqual(compiler._import_files_cache, {})
        self.assertEqual(set(compiler._mtime_cache), {main_less, main_css})

        self.assertSpyCallCount(getmtime_spy, 2)
        self.assertSpyCalledWith(getmtime_spy, main_less)
        self.assertSpyCalledWith(getmtime_spy, main_css)

        self.assertSpyNotCalled(get_import_files_spy)

    def test_is_outdated_with_no_dest_file(self):
        """Testing LessCompiler.is_outdated with missing destination file"""
        compiler = self.compiler
        dep3_less = self.dep3_less
        dep3_css = self.dep3_css

        getmtime_spy = self.spy_on(os.path.getmtime)
        get_import_files_spy = self.spy_on(compiler._get_import_files)

        self.assertFalse(os.path.exists(dep3_css))
        self.assertTrue(compiler.is_outdated(dep3_less, dep3_css))
        self.assertEqual(compiler._import_files_cache, {})

        self.assertEqual(set(compiler._mtime_cache), {dep3_css})
        self.assertSpyCallCount(getmtime_spy, 1)
        self.assertSpyCalledWith(getmtime_spy, dep3_css)

        self.assertSpyNotCalled(get_import_files_spy)

    def test_is_outdated_with_outdated_dep(self):
        """Testing LessCompiler.is_outdated with outdated import file"""
        compiler = self.compiler
        main_less = self.main_less
        main_css = self.main_css

        getmtime_spy = self.spy_on(os.path.getmtime)
        get_import_files_spy = self.spy_on(compiler._get_import_files)

        self.assertFalse(compiler.is_outdated(main_less, main_css))
        self.assertEqual(
            compiler._import_files_cache,
            {
                main_less: set(self.dep_filenames),
            })

        self.assertEqual(set(compiler._mtime_cache),
                         set(self.all_filenames))

        self.assertSpyCallCount(getmtime_spy, 5)

        for filename in self.all_filenames:
            self.assertSpyCalledWith(getmtime_spy, filename)

        self.assertSpyCallCount(get_import_files_spy, 1)
        self.assertSpyCalledWith(get_import_files_spy, main_less)

        # Update the timestamp and try again.
        self._touch(self.dep3_less)

        getmtime_spy.reset_calls()
        get_import_files_spy.reset_calls()

        self.assertTrue(compiler.is_outdated(main_less, main_css))
        self.assertEqual(compiler._import_files_cache, {})
        self.assertEqual(set(compiler._mtime_cache),
                         set(self.all_filenames))

        self.assertSpyCallCount(getmtime_spy, 5)

        for filename in self.all_filenames:
            self.assertSpyCalledWith(getmtime_spy, filename)

        self.assertSpyCallCount(get_import_files_spy, 1)
        self.assertSpyCalledWith(get_import_files_spy, main_less)

        # Make sure mtimes stay cached. We will have to re-import again due
        # to a dependency cache being invalidated.
        getmtime_spy.reset_calls()
        get_import_files_spy.reset_calls()

        self.assertTrue(compiler.is_outdated(main_less, main_css))
        self.assertEqual(compiler._import_files_cache, {})
        self.assertEqual(set(compiler._mtime_cache),
                         set(self.all_filenames))

        self.assertSpyNotCalled(getmtime_spy)

        self.assertSpyCallCount(get_import_files_spy, 1)
        self.assertSpyCalledWith(get_import_files_spy, main_less)

    def _touch(self, filename):
        """Update the timestamp on a file.

        Args:
            filename (unicode):
                The filename of the file to update.
        """
        os.utime(filename, None)
        self.compiler.clear_caches()



from django.views.decorators.cache import never_cache


def never_cache_patterns(*args):
    """Prevent any included URLs from being cached by the browser.

    It's sometimes desirable not to allow browser caching for a set of URLs.
    Any URLs passed in will have the
    :py:func:`~django.views.decorators.cache.never_cache` decorator applied.

    Args:
        *args (tuple):
            The URL arguments to pass to the function.

    Returns:
        list:
        A list of URL patterns.
    """
    pattern_list = []

    for t in args:
        t.callback = never_cache(t.callback)
        pattern_list.append(t)

    return pattern_list



"""Utilities for going between SiteConfiguration and Django settings."""

from __future__ import annotations

from typing import Callable, Dict, Optional, Union

from django.conf import LazySettings, settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.cache import DEFAULT_CACHE_ALIAS
from django.utils import timezone
from typing_extensions import NotRequired, TypeAlias, TypedDict

from djblets.cache.backend_compat import normalize_cache_backend
from djblets.cache.forwarding_backend import (DEFAULT_FORWARD_CACHE_ALIAS,
                                              ForwardingCacheBackend)
from djblets.siteconfig.models import (SiteConfiguration,
                                       SiteConfigurationSettings,
                                       SiteConfigurationSettingsValue)


class SiteConfigurationMappingDynamicValue(TypedDict):
    """A dynamic value for serializing/deserializing siteconfig values.

    Version Added:
        3.1
    """

    #: The Django settings key to set.
    #:
    #: Type:
    #:     str
    key: str

    #: A function for deserializing data from siteconfig settings.
    #:
    #: Type:
    #:     callable
    deserialize_func: NotRequired[Callable[[SiteConfigurationSettingsValue],
                                           SiteConfigurationSettingsValue]]

    #: A function for setting data in Django.
    #:
    #: Type:
    #:     callable
    setter: NotRequired[Callable[[LazySettings, str, object], None]]


#: A dictionary mapping siteconfig keys to Django serialization information.
#:
#: Version Added:
#:     3.1
SiteConfigurationSettingsMap: TypeAlias = \
    Dict[str, Union[str, SiteConfigurationMappingDynamicValue]]


def _set_cache_backend(
    settings: LazySettings,
    key: str,
    value: SiteConfigurationSettingsValue,
) -> None:
    """Set the default cache backend.

    This will set the default cache backend to use Djblets's
    :py:class:`~djblets.cache.forwarding_backend.ForwardingCacheBackend`,
    and use that to forward on to the provided cache backend.

    The new cache settings will be instantly usable by the application
    without having to restart.

    Args:
        settings (django.conf.LazySettings):
            The Django settings object.

        key (str, unused):
            The settings key (``CACHES``).

        value (object):
            The cache backend settings. This may be a legacy URI or a
            dictionary containing cache backend information.
    """
    settings.CACHES.update({
        DEFAULT_FORWARD_CACHE_ALIAS: (
            normalize_cache_backend(value, DEFAULT_FORWARD_CACHE_ALIAS) or
            normalize_cache_backend(value)),
        DEFAULT_CACHE_ALIAS: {
            'BACKEND': '%s.%s' % (ForwardingCacheBackend.__module__,
                                  ForwardingCacheBackend.__name__),
            'LOCATION': DEFAULT_FORWARD_CACHE_ALIAS,
        },
    })

    from django.core.cache import cache

    if isinstance(cache, ForwardingCacheBackend):
        cache.reset_backend()


def _set_static_url(
    settings: LazySettings,
    key: str,
    value: SiteConfigurationSettingsValue,
) -> None:
    """Set the URL for static media.

    Args:
        settings (django.conf.LazySettings):
            The Django settings object.

        key (str, unused):
            The settings key (``STATIC_URL``).

        value (object):
            The new static URL.
    """
    settings.STATIC_URL = value
    staticfiles_storage.base_url = value


def _set_timezone(
    settings: LazySettings,
    key: str,
    value: SiteConfigurationSettingsValue,
) -> None:
    """Set the server's timezone.

    Args:
        settings (django.conf.LazySettings):
            The Django settings object.

        key (str, unused):
            The settings key (``TIME_ZONE``).

        value (object):
            The new timezone.
    """
    assert isinstance(value, str)

    settings.TIME_ZONE = value

    # Internally, Django will also set os.environ['TZ'] to this value
    # and call time.tzset() when initially loading settings. We don't do
    # that, because it can have consequences.
    #
    # You can think of the timezone being set initially by Django as being
    # the core timezone that will be used for anything outside of a request.
    # What we set here is the timezone that Django will use in its own
    # timezone-related functions (for DateTimeFields and the like).
    #
    # That does mean that time.localtime and other functions will not
    # produce reliable dates. However, we need to ensure that any date/time
    # code is timezone-aware anyway, and works with our setting.
    #
    # To see how using os.environ['TZ'] would cause us problems, read
    # http://blog.chipx86.com/2013/01/26/weird-bugs-django-timezones-and-importing-from-eggs/
    timezone.activate(settings.TIME_ZONE)


locale_settings_map: SiteConfigurationSettingsMap = {
    'locale_timezone':             {'key': 'TIME_ZONE',
                                    'deserialize_func': str,
                                    'setter': _set_timezone},
    'locale_language_code':        'LANGUAGE_CODE',
    'locale_date_format':          'DATE_FORMAT',
    'locale_datetime_format':      'DATETIME_FORMAT',
    'locale_default_charset':      {'key': 'DEFAULT_CHARSET',
                                    'deserialize_func': str},
    'locale_language_code':        'LANGUAGE_CODE',
    'locale_month_day_format':     'MONTH_DAY_FORMAT',
    'locale_time_format':          'TIME_FORMAT',
    'locale_year_month_format':    'YEAR_MONTH_FORMAT',
}

mail_settings_map: SiteConfigurationSettingsMap = {
    'mail_server_address':         'SERVER_EMAIL',
    'mail_default_from':           'DEFAULT_FROM_EMAIL',
    'mail_host':                   'EMAIL_HOST',
    'mail_port':                   'EMAIL_PORT',
    'mail_host_user':              {'key': 'EMAIL_HOST_USER',
                                    'deserialize_func': str},
    'mail_host_password':          {'key': 'EMAIL_HOST_PASSWORD',
                                    'deserialize_func': str},
    'mail_use_tls':                'EMAIL_USE_TLS',
}

site_settings_map: SiteConfigurationSettingsMap = {
    'site_media_root':             'MEDIA_ROOT',
    'site_media_url':              'MEDIA_URL',
    'site_static_root':            'STATIC_ROOT',
    'site_static_url':             {'key': 'STATIC_URL',
                                    'setter': _set_static_url},
    'site_prepend_www':            'PREPEND_WWW',
    'site_upload_temp_dir':        'FILE_UPLOAD_TEMP_DIR',
    'site_upload_max_memory_size': 'FILE_UPLOAD_MAX_MEMORY_SIZE',
}

cache_settings_map: SiteConfigurationSettingsMap = {
    'cache_backend':               {'key': 'CACHES',
                                    'setter': _set_cache_backend},
    'cache_expiration_time':       'CACHE_EXPIRATION_TIME',
}


# Don't build unless we need it.
_django_settings_map: SiteConfigurationSettingsMap = {}


def get_django_settings_map() -> SiteConfigurationSettingsMap:
    """Return a map of customizable Django settings.

    These maps are passed to other functions, like :py:func:`generate_defaults`
    and :py:func:`apply_django_settings`. Consumers can make their own
    settings map based on this with additional settings they want to provide.

    Each entry maps a siteconfig settings key to either a Django settings key
    or detailed information on serializing/deserializing a Django setting.

    An entry with detailed information represents that as a dictionary
    containing the following fields:

    Keys:
        key (str):
            The Django settings key.

        deserialize_func (callable, optional):
            A function taking the value from the siteconfig and returning a
            value usable in Django settings.

        setter (callable, optional):
            A function taking the Django settings object, the Django settings
            key, and the new value from the siteconfig (after going through
            ``deserialize_func``, if provided). This will set the value in
            Django's settings object.

    Returns:
        dict:
        The resulting settings map. This is generated once and cached for
        future calls.
    """
    if not _django_settings_map:
        _django_settings_map.update(locale_settings_map)
        _django_settings_map.update(mail_settings_map)
        _django_settings_map.update(site_settings_map)
        _django_settings_map.update(cache_settings_map)

    return _django_settings_map


def generate_defaults(
    settings_map: SiteConfigurationSettingsMap,
) -> SiteConfigurationSettings:
    """Return a dictionary of siteconfig defaults for Django settings.

    This will iterate through the provided settings map and return a dictionary
    mapping a siteconfig settings key to the value from Django's settings (if
    one is found).

    Args:
        settings_map (dict):
            A settings map, generated from :py:func:`get_django_settings_map`
            or a similar function.

    Returns:
        dict:
        A dictionary of siteconfig defaults.
    """
    defaults: SiteConfigurationSettings = {}

    for siteconfig_key, setting_data in settings_map.items():
        if isinstance(setting_data, dict):
            setting_key = setting_data['key']
        else:
            setting_key = setting_data

        if hasattr(settings, setting_key):
            defaults[siteconfig_key] = getattr(settings, setting_key)

    return defaults


def get_locale_defaults() -> SiteConfigurationSettings:
    """Return a dictionary of siteconfig defaults for Django locale settings.

    The generated defaults are specific to the values in
    :py:data:`locale_settings_map`.

    Returns:
        dict:
        A dictionary of siteconfig defaults.
    """
    return generate_defaults(locale_settings_map)


def get_mail_defaults() -> SiteConfigurationSettings:
    """Return a dictionary of siteconfig defaults for Django e-mail settings.

    The generated defaults are specific to the values in
    :py:data:`mail_settings_map`.

    Returns:
        dict:
        A dictionary of siteconfig defaults.
    """
    return generate_defaults(mail_settings_map)


def get_site_defaults() -> SiteConfigurationSettings:
    """Return a dictionary of siteconfig defaults for Django site settings.

    The generated defaults are specific to the values in
    :py:data:`site_settings_map`.

    Returns:
        dict:
        A dictionary of siteconfig defaults.
    """
    return generate_defaults(site_settings_map)


def get_cache_defaults() -> SiteConfigurationSettings:
    """Return a dictionary of siteconfig defaults for Django caching settings.

    The generated defaults are specific to the values in
    :py:data:`cache_settings_map`.

    Returns:
        dict:
        A dictionary of siteconfig defaults.
    """
    return generate_defaults(cache_settings_map)


def get_django_defaults() -> SiteConfigurationSettings:
    """Return a dictionary of siteconfig defaults for Django settings.

    The generated defaults are specific to the values returned by
    :py:func:`get_django_settings_map`.

    Returns:
        dict:
        A dictionary of siteconfig defaults.
    """
    return generate_defaults(get_django_settings_map())


def apply_django_settings(
    siteconfig: SiteConfiguration,
    settings_map: Optional[SiteConfigurationSettingsMap] = None,
) -> None:
    """Apply Django settings stored in the site configuration.

    This takes a siteconfiguration storing Django settings and a settings map,
    applying each of the settings to Django. Setting will generally be stored
    in the Django settings object, but some settings will be specially applied
    based on their rules in the settings map.

    Args:
        siteconfig (djblets.siteconfig.models.SiteConfiguration):
            The site configuration containing the Django settings to apply.

        settings_map (dict, optional):
            A map of siteconfig keys to Django settings information. See
            :py:func:`get_django_settings_map` for details.

            If not provided, the result of :py:func:`get_django_settings_map`
            will be used.
    """
    if settings_map is None:
        settings_map = get_django_settings_map()

    for key, setting_data in settings_map.items():
        if key in siteconfig.settings:
            value = siteconfig.get(key)
            setter: Callable = setattr

            if isinstance(setting_data, dict):
                setting_key = setting_data['key']

                if 'setter' in setting_data:
                    setter = setting_data['setter']

                if ('deserialize_func' in setting_data and
                    callable(setting_data['deserialize_func'])):
                    value = setting_data['deserialize_func'](value)
            else:
                setting_key = setting_data

            setter(settings, setting_key, value)



"""Components for creating customizable datagrids from database data.

Datagrids are used to display a table-based view of data from a database,
complete with pagination, batch selection, sorting, and flexible column
rendering.

Datagrids have one or more :py:class:`Column` subclasses associated, which will
render the data. The datagrid may display a subset of the rendered columns,
and users can choose which of those columns they want displayed, and in which
order.

There are two main types of datagrids:

* :py:class:`DataGrid` is the base class for a datagrid, and will display
  the data with standard numerical page-based pagination.

* :py:class:`AlphanumericDataGrid` is similar, but uses a more specific
  paginator that allows the user to paginate by the first letter/number/symbol
  of the data in a given field. This is useful for lists of users, for
  example.

All datagrids are meant to be subclassed.
"""

import logging
import re
import string
import traceback

import pytz
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.paginator import InvalidPage, Paginator
from django.http import Http404, HttpResponse
from django.template.defaultfilters import date, timesince
from django.template.loader import get_template, render_to_string
from django.utils.cache import patch_cache_control
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from djblets.deprecation import RemovedInDjblets40Warning
from djblets.template.context import get_default_template_context_processors
from djblets.util.decorators import cached_property
from djblets.util.http import get_url_params_except


logger = logging.getLogger(__name__)


# Registration of all datagrid classes to columns.
_column_registry = {}


class Column(object):
    """A column in a datagrid.

    The column is the primary component of the datagrid. It is used to
    display not only the column header but the HTML for the cell as well.

    Columns can be tied to database fields and can be used for sorting.
    Not all columns have to allow for this, though.

    Columns can have an image, text, or both in the column header. The
    contents of the cells can be instructed to link to the object on the
    row or the data in the cell.

    If a Column defines an :py:attr:`image_class`, then it will be assumed that
    the class represents an icon, perhaps as part of a spritesheet, and will
    display it in a ``<div>``. An :py:attr:`image_url` cannot also be defined.

    Attributes:
        cell_template (unicode):
            The path to a template. If this is not None, this will override
            the default cell_template for the DataGrid the column is in.
    """

    #: Descending sort order for columns.
    SORT_DESCENDING = 0

    #: Ascending sort order for columns.
    SORT_ASCENDING = 1

    def __init__(self, label=None, id=None, detailed_label=None,
                 detailed_label_html=None, field_name=None, db_field=None,
                 image_url=None, image_class=None, image_width=None,
                 image_height=None, image_alt="", shrink=False, expand=False,
                 sortable=False,
                 default_sort_dir=SORT_DESCENDING, link=False,
                 link_func=None, link_css_class=None, cell_clickable=False,
                 css_class=""):
        """Initialize the column.

        When initializing a column as part of a :py:class:`DataGrid` subclass,
        a number of options can be provided.

        Args:
            id (unicode, optional):
                The unique ID of the column on the datagrid.

            label (unicode, optional):
                The label to show in the column header.

            detailed_label (unicode, optional):
                A detailed label to display in the column customization
                menu. Defaults to ``label``.

            detailed_label_html (unicode, optional):
                A detailed label in HTML form to display in the column
                customization menu. This takes precedence over
                ``detailed_label``.

            field_name (unicode, optional):
                The name of the field on the model containing the data to
                render.

            db_field (unicode, optional):
                The name of the database field containing the field used
                for sorting. Defaults to ``field_name``.

            image_url (unicode, optional):
                The URL to the image used in the header and navigation menu.
                This cannot be used with ``image_class``.

            image_class (unicode, optional):
                The CSS class of a spritesheet icon to use in the header
                and navigation menu. This cannot be used with ``image_url``.

            image_width (int, optional):
                The width of the image.

            image_height (int, optional):
                The height of the image.

            image_alt (unicode, optional):
                The alt text for the image.

            shrink (bool, optional):
                If ``True``, the column's width will be calculated to its
                minimum size.

            expand (bool, optional):
                If ``True``, the column's width will be calculated to its
                maximum size. If there are other expanded columns, they'll
                share the available width equally.

            sortable (bool, optional):
                If ``True``, the column can be sorted. This requires a
                ``db_field`` that allows for sorting.

            default_sort_dir (int, optional):
                The default sorting direction when the user activates sorting.
                Either :py:attr:`SORT_DESCENDING`
                or :py:attr:`SORT_ASCENDING`.

            link (bool, optional):
                If ``True``, the contents will be linked to the URL
                returned by ``link_func`` or
                :py:meth:`DataGrid.link_to_object`.

            link_func (callable, optional):
                Optional function that returns a URL for the link.

            link_css_class (unicode or callable, optional):
                The CSS class or classes to define on ``<a>`` for the link
                for the cell, if setting ``link=True``. This can be a
                function returning the classes.

            cell_clickable (bool, optional):
                If ``True``, clicking anywhere on the cell will navigate to
                the URL defined, if any.

            css_class (unicode, optional):
                The CSS class or classes to define on the cell. This can be
                a function returning the classes.
        """
        assert not (image_class and image_url)

        self.id = id
        self.field_name = field_name
        self.db_field = db_field or field_name
        self.label = label
        self.detailed_label = detailed_label or self.label
        self.detailed_label_html = detailed_label_html or self.detailed_label
        self.image_url = image_url
        self.image_class = image_class
        self.image_width = image_width
        self.image_height = image_height
        self.image_alt = image_alt
        self.shrink = shrink
        self.expand = expand
        self.sortable = sortable
        self.default_sort_dir = default_sort_dir
        self.cell_clickable = False
        self.link = link
        self.link_func = (
            link_func or
            (lambda state, x, y: state.datagrid.link_to_object(state, x, y)))
        self.link_css_class = link_css_class
        self.css_class = css_class

        self.cell_template = None

    @cached_property
    def cell_template_obj(self):
        """Return the cell template, if it exists."""
        if self.cell_template:
            return get_template(self.cell_template)

        return None

    def setup_state(self, state):
        """Set up any state that may be needed for the column.

        This is called once per column per datagrid instance.

        By default, no additional state is set up. Subclasses can override
        this to set any variables they may need.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.
        """
        pass

    def get_sort_field(self, state):
        """Return the field used for sorting this column.

        By default, this uses the provided db_field.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.

        Returns:
            unicode:
                The field on the model used for sorting. Defaults
                to ``db_field``.
        """
        return self.db_field

    def get_toggle_url(self, state):
        """Return a URL to toggle this column's visibility.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.

        Returns:
            unicode: The URL used to toggle column visibility.
        """
        columns = [column.id for column in state.datagrid.columns]

        if state.active:
            try:
                columns.remove(self.id)
            except ValueError:
                pass
        else:
            columns.append(self.id)

        url_params = get_url_params_except(state.datagrid.request.GET,
                                           'columns')
        if url_params:
            url_params = url_params + '&'

        return "?%scolumns=%s" % (url_params, ",".join(columns))

    def get_header(self, state):
        """Render the header for the column.

        The column header will include the current sort indicator, if it
        belongs in the sort list. It will also be made clickable in order
        to modify the sort order appropriately, if sortable.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.

        Returns:
            unicode: The HTML for the header.
        """
        datagrid = state.datagrid
        in_sort = False
        sort_direction = self.SORT_DESCENDING
        sort_primary = False
        sort_url = ""
        unsort_url = ""

        if self.sortable:
            sort_list = list(datagrid.sort_list)

            if sort_list:
                rev_column_id = "-%s" % self.id
                new_column_id = self.id
                cur_column_id = ""

                if self.id in sort_list:
                    # This column is currently being sorted in
                    # ascending order.
                    sort_direction = self.SORT_ASCENDING
                    cur_column_id = self.id
                    new_column_id = rev_column_id
                elif rev_column_id in sort_list:
                    # This column is currently being sorted in
                    # descending order.
                    sort_direction = self.SORT_DESCENDING
                    cur_column_id = rev_column_id
                    new_column_id = self.id

                if cur_column_id:
                    in_sort = True
                    sort_primary = (sort_list[0] == cur_column_id)

                    if not sort_primary:
                        # If this is not the primary column, we want to keep
                        # the sort order intact.
                        new_column_id = cur_column_id

                    # Remove this column from the current location in the list
                    # so we can move it to the front of the list.
                    sort_list.remove(cur_column_id)

                # Insert the column name into the beginning of the sort list.
                sort_list.insert(0, new_column_id)
            else:
                # There's no sort list to begin with. Make this column
                # the only entry.
                sort_list = [self.id]

            # We can only support two entries in the sort list, so truncate
            # this.
            del(sort_list[2:])

            url_params = get_url_params_except(
                datagrid.request.GET,
                "sort", "datagrid-id", "gridonly", "columns")
            if url_params:
                url_params = url_params + '&'

            url_prefix = "?%ssort=" % url_params
            unsort_url = url_prefix + ','.join(sort_list[1:])
            sort_url = url_prefix + ','.join(sort_list)

        return datagrid.column_header_template_obj.render(context={
            'column': self,
            'column_state': state,
            'in_sort': in_sort,
            'sort_ascending': sort_direction == self.SORT_ASCENDING,
            'sort_primary': sort_primary,
            'sort_url': sort_url,
            'unsort_url': unsort_url,
        })

    def collect_objects(self, state, object_list):
        """Iterate through the objects and builds a cache of data to display.

        This optimizes the fetching of data in the grid by grabbing all the
        IDs of related objects that will be queried for rendering, loading
        them all at once, and populating the cache.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.

            object_list (list):
                The list of objects being rendered on the datagrid.
        """
        id_field = '%s_id' % self.field_name
        ids = set()
        model = None

        for obj in object_list:
            if not hasattr(obj, id_field):
                # This isn't the field type you're looking for.
                return

            id_value = getattr(obj, id_field)

            if id_value is None:
                continue

            ids.add(id_value)

            if not model:
                field = getattr(obj.__class__, self.field_name).field

                try:
                    model = field.rel.to
                except AttributeError:
                    # No idea what this is. Bail.
                    return

        if model and ids:
            for obj in model.objects.filter(pk__in=ids):
                state.data_cache[obj.pk] = obj

    def render_cell(self, state, obj, render_context):
        """Render the table cell containing column data.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.

            obj (object):
                The object being rendered for this row.

            render_context (Context):
                The shared context used for cell renders.

        Returns:
            unicode: The rendered cell as HTML.
        """
        try:
            rendered_data = self.render_data(state, obj)
        except Exception as e:
            logger.exception('Error when calling render_data for DataGrid '
                             'Column %r: %s',
                             self, e)
            rendered_data = ''

        # We use empty strings instead of None just to keep cache keys small.
        url = ''
        css_class = ''
        link_css_class = ''

        if self.link:
            if self.link_func:
                try:
                    url = self.link_func(state, obj, rendered_data)
                except AttributeError:
                    pass
            elif render_context:
                url = render_context.get('_datagrid_object_url')

        if self.css_class:
            if callable(self.css_class):
                css_class = self.css_class(obj)
            else:
                css_class = self.css_class

        if self.link_css_class:
            if callable(self.link_css_class):
                link_css_class = self.link_css_class(obj)
            else:
                link_css_class = self.link_css_class

        key = '%s:%s:%s:%s:%s' % (state.last, rendered_data, url, css_class,
                                  link_css_class)

        if key not in state.cell_render_cache:
            if url:
                css_class = '%s has-link' % css_class

            ctx = {}

            if render_context:
                ctx.update(render_context)

            ctx.update({
                'column': self,
                'column_state': state,
                'css_class': css_class.strip(),
                'link_css_class': link_css_class,
                'url': url,
                'data': mark_safe(rendered_data)
            })

            template = self.cell_template_obj

            if template is None:
                template = state.datagrid.cell_template_obj

            state.cell_render_cache[key] = template.render(ctx)

        return state.cell_render_cache[key]

    def render_data(self, state, obj):
        """Render the column data within the cell.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.

            obj (object):
                The object being rendered for this row.

        Returns:
            unicode: The rendered data as HTML.
        """
        id_field = '%s_id' % self.field_name

        # Look for this directly so that we don't end up fetching the
        # data for the object.
        if id_field in obj.__dict__:
            pk = obj.__dict__[id_field]

            if pk in state.data_cache:
                return state.data_cache[pk]
            else:
                value = getattr(obj, self.field_name)
                state.data_cache[pk] = escape(value)
                return value
        else:
            # Follow . separators like in the django template library
            value = obj
            for field_name in self.field_name.split('.'):
                if field_name:
                    value = getattr(value, field_name)

                    if callable(value):
                        value = value()

            return escape(value)

    def augment_queryset(self, state, queryset):
        """Augment a queryset with new queries.

        Subclasses can override this to extend the queryset to provide
        additional information, usually using queryset.extra(). This must
        return a queryset based on the original queryset.

        This should not restrict the query in any way, or the datagrid may
        not operate properly. It must only add additional data to the
        queryset.

        Args:
            state (StatefulColumn):
                The state for the DataGrid instance.

            queryset (QuerySet):
                The queryset to augment.

        Returns:
            QuerySet: The resulting QuerySet.
        """
        return queryset


class StatefulColumn(object):
    """A stateful wrapper for a Column instance.

    Columns must be stateless, as they are shared across all instances of
    a particular DataGrid. However, some state is needed for columns, such
    as their widths or active status.

    StatefulColumn wraps a :py:class:`Column` instance and provides state
    storage, and also provides a convenient way to call methods on a Column and
    pass the state.

    Attributes owned by the Column can be accessed directly through the
    StatefulColumn.

    Likewise, any functions owned by the Column can be accessed as well.
    The function will be invoked with this StatefulColumn as the first
    parameter passed.
    """

    def __init__(self, datagrid, column):
        """Initialize the column state.

        Args:
            datagrid (DataGrid):
                The DataGrid instance owning this column state.

            column (Column):
                The column instance this state is associated with.
        """
        self.datagrid = datagrid
        self.column = column
        self.active = False
        self.last = False
        self.width = 0
        self.data_cache = {}
        self.cell_render_cache = {}

        try:
            column.setup_state(self)
        except Exception as e:
            logger.exception('Error when calling setup_state for DataGrid '
                             'Column %r: %s',
                             self.column, e)

    @property
    def toggle_url(self):
        """The visibility toggle URL of the column.

        This is a convenience used by templates to call
        :py:meth:`Column.get_toggle_url` with the current state.
        """
        return self.column.get_toggle_url(self)

    @property
    def header(self):
        """The header of the column.

        This is a convenience used by templates to call
        :py:meth:`Column.get_header` with the current state.
        """
        return self.column.get_header(self)

    def __getattr__(self, name):
        """Returns an attribute from the parent Column.

        This is called when accessing an attribute not found directly on
        StatefulColumn. The attribute will be fetched from the Column
        (if it exists there).

        In the case of accessing a function, a wrapper will be returned
        that will automatically pass this StatefulColumn instance as the
        first parameter.

        Args:
            name (unicode): The attribute to fetch from the column.

        Returns:
            The attribute value from the column.
        """
        result = getattr(self.column, name)

        if callable(result):
            return lambda *args, **kwargs: result(self, *args, **kwargs)

        return result


class CheckboxColumn(Column):
    """A column that renders a checkbox.

    The :py:meth:`is_selectable` and :py:meth:`is_selected` functions can be
    overridden to control whether a checkbox is displayed in a row and whether
    that checkbox is initially checked.

    The checkboxes have a ``data-object-id`` attribute that contains the ID of
    the object that row represents. This allows the JavaScript code to
    determine which rows have been checked, and operate on that
    accordingly.

    The checkboxes also have a ``data-checkbox-name`` attribute that
    contains the value passed in to the ``checkbox_name`` parameter of its
    constructor.
    """

    def __init__(self, checkbox_name='select', shrink=True,
                 show_checkbox_header=True,
                 detailed_label=_('Select Rows'),
                 *args, **kwargs):
        """Initialize the column.

        Args:
            checkbox_name (unicode):
                The name set in ``data-checkbox-name``.

            shrink (bool):
                If ``True``, the column's width will be calculated to its
                minimum size.

            show_checkbox_header (bool):
                If ``True``, a checkbox will be used for the column header.
        """
        super(CheckboxColumn, self).__init__(
            shrink=shrink,
            label=format_html(
                '<input class="datagrid-header-checkbox"'
                ' type="checkbox" data-checkbox-name="{0}" />',
                checkbox_name),
            detailed_label=detailed_label,
            detailed_label_html=format_html(
                '<input type="checkbox" /> {0}',
                detailed_label),
            *args, **kwargs)

        self.show_checkbox_header = show_checkbox_header
        self.checkbox_name = checkbox_name
        self.cell_template = 'datagrid/cell_no_link.html'

    def render_data(self, state, obj):
        if self.is_selectable(state, obj):
            checked = ''

            if self.is_selected(state, obj):
                checked = mark_safe('checked="true"')

            return format_html('<input type="checkbox" data-object-id="{0}" '
                               'data-checkbox-name="{1}" {2} />',
                               obj.pk, self.checkbox_name, checked)
        else:
            return ''

    def is_selectable(self, state, obj):
        """Returns whether an object can be selected.

        If this returns False, no checkbox will be rendered for this item.
        """
        return True

    def is_selected(self, state, obj):
        """Returns whether an object has been selected.

        If this returns True, the checkbox will be checked.
        """
        return False


class DateTimeColumn(Column):
    """A column that renders a date or time."""

    def __init__(self, label, format=None, sortable=True,
                 timezone=pytz.utc, *args, **kwargs):
        super(DateTimeColumn, self).__init__(label, sortable=sortable,
                                             *args, **kwargs)
        self.format = format
        self.timezone = timezone

    def render_data(self, state, obj):
        # If the datetime object is tz aware, conver it to local time
        datetime = getattr(obj, self.field_name)
        if settings.USE_TZ:
            datetime = pytz.utc.normalize(datetime).\
                astimezone(self.timezone)

        return date(datetime, self.format)


class DateTimeSinceColumn(Column):
    """A column that renders a date or time relative to now."""

    def __init__(self, label, sortable=True, timezone=pytz.utc,
                 *args, **kwargs):
        super(DateTimeSinceColumn, self).__init__(label, sortable=sortable,
                                                  *args, **kwargs)

    def render_data(self, state, obj):
        return _("%s ago") % timesince(getattr(obj, self.field_name))


class DataGrid(object):
    """A paginated table of data based on queries from a database.

    A datagriad represents a list of objects, sorted and organized by
    columns. The sort order and column lists can be customized. allowing
    users to view this data however they prefer.

    This is meant to be subclassed for specific uses. The subclasses are
    responsible for defining one or more column types. It can also set
    one or more of the following optional variables:

    Attributes:
        title (unicode):
            The title of the grid.

        profile_sort_field (unicode):
            The variable name in the user profile where the sort order can be
            loaded and saved.

        profile_columns_field (unicode):
            The variable name in the user profile where the columns list can be
            loaded and saved.

        paginate_by (int):
            The number of items to show on each page of the grid. The default
            is 50.

        paginate_orphans (int):
            If this number of objects or fewer are on the last page, it will be
            rolled into the previous page. The default is 3.

        page (int):
            The page to display. If this is not specified, the ``?page=``
            variable passed in the URL will be used, or 1 if that is not
            specified.

        listview_template (unicode):
            The template used to render the list view. The default is
            :file:`datagrid/listview.html`.

        column_header_template (unicode):
            The template used to render each column header. The default is
            :file:`datagrid/column_header.html`.

        cell_template (unicode):
            The template used to render a cell of data. The default is
            :file:`datagrid/cell.html`.

        optimize_sorts (bool):
            Whether or not to optimize queries when using multiple sorts. This
            can offer a speed improvement, but may need to be turned off for
            more advanced querysets (such as when using ``extra()``).
            The default is ``True``.
    """

    _columns = None

    @classmethod
    def add_column(cls, column):
        """Add a new column for this datagrid.

        This can be used to add columns to a DataGrid subclass after
        the subclass has already been defined.

        The column added must have a unique ID already set.

        Args:
            column (Column):
                The column to add.
        """
        cls._populate_columns()

        if not column.id:
            raise KeyError(
                'Custom datagrid columns must have a unique id attribute.')

        if column.id in _column_registry[cls]:
            raise KeyError('"%s" is already a registered column for %s'
                           % (column.id, cls.__name__))

        _column_registry[cls][column.id] = column

    @classmethod
    def remove_column(cls, column):
        """Remove a column from this datagrid.

        This can be used to remove columns previously added through
        :py:meth:`add_column`.

        Args:
            column (Column):
                The column to remove.
        """
        cls._populate_columns()

        try:
            del _column_registry[cls][column.id]
        except KeyError:
            raise KeyError('"%s" is not a registered column for %s'
                           % (column.id, cls.__name__))

    @classmethod
    def get_column(cls, column_id):
        """Return the column with the given ID.

        If not found, this will return None.

        Args:
            column_id (int):
                The index of the column to return.

        Returns:
            Column: The resulting column at the given index.
        """
        cls._populate_columns()

        return _column_registry[cls].get(column_id)

    @classmethod
    def get_columns(cls):
        """Return the list of registered columns for this datagrid.

        Returns:
            list of Column: The list of columns registered on this datagrid.
        """
        cls._populate_columns()

        return _column_registry[cls].values()

    @classmethod
    def _populate_columns(cls):
        """Populate the default list of columns for the datagrid.

        The default list contains all columns added in the class definition.
        """
        if cls not in _column_registry:
            _column_registry[cls] = {}

            for key in dir(cls):
                column = getattr(cls, key)

                if isinstance(column, Column):
                    column.id = key

                    if not column.field_name:
                        column.field_name = column.id

                    if not column.db_field:
                        column.db_field = column.field_name

                    cls.add_column(column)

    def __init__(self, request, queryset=None, title="", extra_context={},
                 optimize_sorts=True, model=None):
        """Initialize the datagrid.

        Args:
            request (HttpRequest):
                The HTTP request from the client.

            queryset (QuerySet):
                A QuerySet returning the objects to render in the grid.

            title (unicode):
                The displayed title of the datagrid.

            extra_context (dict):
                Extra context variables to render on the datagrid template.

            optimize_sorts (bool):
                If ``True``, sorting will be optimized, reducing the
                complexity of the queries. This is the default.

            model (Model):
                The model for the objects in the datagrid. Defaults to the
                model associated with ``queryset``.
        """
        self.request = request
        self.queryset = queryset
        self.rows = []
        self.columns = []
        self.column_map = {}
        self.id_list = []
        self.paginator = None
        self.page = None
        self.sort_list = None
        self.state_loaded = False
        self.page_num = 0
        self.id = None
        self.extra_context = dict(extra_context)
        self.optimize_sorts = optimize_sorts
        self.special_query_args = []
        self._model = model

        if not hasattr(request, "datagrid_count"):
            request.datagrid_count = 0

        self.id = "datagrid-%s" % request.datagrid_count
        request.datagrid_count += 1

        # Customizable variables
        self.title = title
        self.profile_sort_field = None
        self.profile_columns_field = None
        self.paginate_by = 50
        self.paginate_orphans = 3
        self.listview_template = 'datagrid/listview.html'
        self.column_header_template = 'datagrid/column_header.html'
        self.cell_template = 'datagrid/cell.html'
        self.paginator_template = 'datagrid/paginator.html'

    @cached_property
    def cell_template_obj(self):
        """The rendered template used for cells on this datagrid.

        This will only be generated once, and reused for all cells.
        """
        obj = get_template(self.cell_template)

        if not obj:
            logger.error("Unable to load template '%s' for datagrid "
                         "cell. This may be an installation issue.",
                         self.cell_template,
                         request=self.request)

        return obj

    @cached_property
    def column_header_template_obj(self):
        """The rendered template used for column headers on this datagrid.

        This will only be generated once, and reused for all headers.
        """
        obj = get_template(self.column_header_template)

        if not obj:
            logger.error("Unable to load template '%s' for datagrid "
                         "column headers. This may be an installation "
                         "issue.",
                         self.column_header_template,
                         request=self.request)

        return obj

    @property
    def all_columns(self):
        """All columns in the datagrid, sorted by label."""
        return [
            self.get_stateful_column(column)
            for column in sorted(self.get_columns(),
                                 key=lambda x: x.detailed_label)
        ]

    @property
    def model(self):
        """The model representing the objects shown in the grid."""
        if self._model is None:
            return self.queryset.model

        return self._model

    def get_stateful_column(self, column):
        """Return a StatefulColumn for the given Column instance.

        If one has already been created, it will be returned.

        Args:
            column (Column):
                The column associated with the stateful column.

        Returns:
            StatefulColumn: The column state associated with the column.
        """
        if column not in self.column_map:
            self.column_map[column] = StatefulColumn(self, column)

        return self.column_map[column]

    def load_state(self, render_context=None):
        """Load the state of the datagrid.

        This will retrieve the user-specified or previously stored
        sorting order and columns list, as well as any state a subclass
        may need.

        Args:
            render_context (Context):
                Common template variable context to render on the datagrid.
        """
        if self.state_loaded:
            return

        profile_sort_list = None
        profile_columns_list = None
        profile = None
        profile_dirty_fields = []
        profile_dirty_fields_all = False

        # Get the saved settings for this grid in the profile. These will
        # work as defaults and allow us to determine if we need to save
        # the profile.
        if self.request.user.is_authenticated:
            profile = self.get_user_profile()

            if profile:
                if self.profile_sort_field:
                    profile_sort_list = \
                        getattr(profile, self.profile_sort_field, None)

                if self.profile_columns_field:
                    profile_columns_list = \
                        getattr(profile, self.profile_columns_field, None)

        # Figure out the columns we're going to display
        # We're also going to calculate the column widths based on the
        # shrink and expand values.
        colnames = self.request.GET.get('columns', profile_columns_list) or ''

        columns = list(filter(None, [
            self.get_column(colname)
            for colname in colnames.split(',')
        ]))

        if not columns:
            colnames = ','.join(self.default_columns)
            columns = [
                self.get_column(colname)
                for colname in self.default_columns
            ]

        expand_columns = []
        normal_columns = []

        for column_def in columns:
            column = self.get_stateful_column(column_def)
            self.columns.append(column)
            column.active = True

            if column.expand:
                # This column is requesting all remaining space. Save it for
                # later so we can tell how much to give it. Each expanded
                # column will count as two normal columns when calculating
                # the normal sized columns.
                expand_columns.append(column)
            elif column.shrink:
                # Make this as small as possible.
                column.width = 0
            else:
                # We'll divide the column widths equally after we've built
                # up the lists of expanded and normal sized columns.
                normal_columns.append(column)

        self.columns[-1].last = True

        # Try to figure out the column widths for each column.
        # We'll start with the normal sized columns.
        total_pct = 100

        # Each expanded column counts as two normal columns.
        normal_column_width = total_pct / (len(self.columns) +
                                           len(expand_columns))

        for column in normal_columns:
            column.width = normal_column_width
            total_pct -= normal_column_width

        if len(expand_columns) > 0:
            expanded_column_width = total_pct / len(expand_columns)
        else:
            expanded_column_width = 0

        for column in expand_columns:
            column.width = expanded_column_width

        # Now get the sorting order for the columns.
        sort_str = self.request.GET.get('sort', profile_sort_list)

        if sort_str:
            self.sort_list = []

            for sort_item in sort_str.split(','):
                if sort_item[0] == '-':
                    base_sort_item = sort_item[1:]
                else:
                    base_sort_item = sort_item

                column = self.get_column(base_sort_item)

                if column and column.sortable:
                    self.sort_list.append(sort_item)

        else:
            self.sort_list = self.default_sort
            sort_str = ",".join(self.sort_list)

        # A subclass might have some work to do for loading and saving
        # as well.
        load_state_result = self.load_extra_state(profile)

        if isinstance(load_state_result, list):
            profile_dirty_fields += load_state_result
        else:
            RemovedInDjblets40Warning.warn(
                '%s.load_extra_state() returned a %s, but must return a '
                'list of field names to save (or an empty list). This will '
                'be required in Djblets 4.0.'
                % (type(self).__name__, type(load_state_result)))

            if load_state_result is True:
                profile_dirty_fields_all = True

        # Now that we have all that, figure out if we need to save new
        # settings back to the profile.
        if profile:
            if (self.profile_columns_field and
                colnames != profile_columns_list):
                setattr(profile, self.profile_columns_field, colnames)
                profile_dirty_fields.append(self.profile_columns_field)

            if self.profile_sort_field and sort_str != profile_sort_list:
                setattr(profile, self.profile_sort_field, sort_str)
                profile_dirty_fields.append(self.profile_sort_field)

            if profile_dirty_fields_all:
                # This can be removed in Djblets 4.
                profile.save()
            elif profile_dirty_fields:
                profile.save(update_fields=profile_dirty_fields)

        self.state_loaded = True

        # Fetch the list of objects and have it ready.
        self.precompute_objects(render_context)

    def get_user_profile(self):
        """Return the object, if any, to use for the user profile state.

        Returns:
            The object, if any, used to store and retrieve persistent
            profile state for the datagrid.
        """
        if hasattr(self.request.user, 'get_profile'):
            try:
                return self.request.user.get_profile()
            except ObjectDoesNotExist:
                pass

        return None

    def load_extra_state(self, profile):
        """Load any extra state needed for this grid.

        This is used by subclasses that may have additional data to load
        and save.

        Version Changed:
            3.0:
            This should now return a list of field names to save in
            ``profile``. Any other result is deprecated and will no longer be
            supported in Djblets 4.0.

        Args:
            profile (django.db.models.Model):
                The profile model instance to load from, if any.

        Returns:
            bool or list of str:
            A list of field names on ``profile`` that have been modified and
            should be saved.

            Djblets 3.0 and older support ``True`` to save the entire object,
            or ``False`` if fields weren't modified. This support will be
            removed in Djblets 4.0.
        """
        return []

    def precompute_objects(self, render_context=None):
        """Pre-compute all objects used to render the datagrid.

        This builds the queryset and stores the list of objects for use in
        rendering the datagrid. It takes into consideration sorting,
        the current page, and augmented queries from columns.

        Args:
            render_context (Context):
                The common template variable context to render on the datagrid,
                provided in the constructor.
        """
        query = self.queryset
        use_select_related = False

        # Generate the actual list of fields we'll be sorting by
        sort_list = []

        for sort_item in self.sort_list:
            if sort_item[0] == '-':
                base_sort_item = sort_item[1:]
                prefix = '-'
            else:
                base_sort_item = sort_item
                prefix = ''

            if sort_item:
                column = self.get_column(base_sort_item)
                if not column:
                    logger.warning('Skipping non-existing sort column "%s"',
                                   base_sort_item,
                                   request=self.request)
                    continue

                elif not column.sortable:
                    logger.warning('Skipping column "%s" which is not '
                                   'sortable',
                                   base_sort_item,
                                   request=self.request.user.username)
                    continue

                stateful_column = self.get_stateful_column(column)

                if stateful_column:
                    try:
                        sort_field = stateful_column.get_sort_field()
                    except Exception as e:
                        logger.exception('Error when calling get_sort_field '
                                         'for DataGrid Column %r: %s',
                                         column, e,
                                         request=self.request)
                        continue

                    if sort_field:
                        sort_list.append(prefix + sort_field)

                    # Lookups spanning tables require that we query from those
                    # tables. In order to keep things simple, we'll just use
                    # select_related so that we don't have to figure out the
                    # table relationships. We only do this if we have a lookup
                    # spanning tables.
                    if '.' in sort_field:
                        use_select_related = True

        if sort_list:
            query = query.order_by(*sort_list)

        query = self.post_process_queryset(query)

        if hasattr(query, 'distinct'):
            query = query.distinct()

        self.paginator = self.build_paginator(query)

        page_num = self.request.GET.get('page', 1)

        # Accept either "last" or a valid page number.
        if page_num == "last":
            page_num = self.paginator.num_pages

        try:
            self.page = self.paginator.page(page_num)
        except InvalidPage:
            raise Http404

        self.id_list = []

        if self.optimize_sorts and len(sort_list) > 0:
            # This can be slow when sorting by multiple columns. If we
            # have multiple items in the sort list, we'll request just the
            # IDs and then fetch the actual details from that.
            if hasattr(self.page.object_list, 'values_list'):
                # This is a standard QuerySet.
                self.id_list = list(self.page.object_list.values_list(
                    'pk', flat=True))
            else:
                # This is something more custom. Perhaps a Haystack
                # SearchQuerySet. It must have a 'pk' or it won't work.
                self.id_list = [int(obj.pk) for obj in self.page.object_list]

            # Make sure to unset the order. We can't meaningfully order these
            # results in the query, as what we really want is to keep it in
            # the order specified in id_list, and we certainly don't want
            # the database to do any special ordering (possibly slowing things
            # down). We'll set the order properly in a minute.
            self.page.object_list = self.post_process_queryset(
                self.model.objects.filter(pk__in=self.id_list).order_by())

        if use_select_related:
            self.page.object_list = \
                self.page.object_list.select_related(depth=1)

        if self.id_list:
            # The database will give us the items in a more or less random
            # order, since it doesn't know to keep it in the order provided by
            # the ID list. This will place the results back in the order we
            # expect.
            index = dict([(id, pos) for (pos, id) in enumerate(self.id_list)])
            object_list = [None] * len(self.id_list)

            for obj in list(self.page.object_list):
                object_list[index[obj.pk]] = obj
        else:
            # Grab the whole list at once. We know it won't be too large,
            # and it will prevent one query per row.
            object_list = list(self.page.object_list)

        for column in self.columns:
            column.collect_objects(object_list)

        if render_context is None:
            render_context = self._build_render_context()

        try:
            self.rows = []

            for obj in object_list:
                if obj is None:
                    continue

                if hasattr(obj, 'get_absolute_url'):
                    obj_url = obj.get_absolute_url()
                else:
                    obj_url = None

                render_context['_datagrid_object_url'] = obj_url

                self.rows.append({
                    'object': obj,
                    'cells': [column.render_cell(obj, render_context)
                              for column in self.columns],
                    'url': obj_url,
                })
        except Exception as e:
            logger.exception('Error when calling render_cell for DataGrid '
                             'Column %r: %s',
                             column, e)

    def post_process_queryset(self, queryset):
        """Add column-specific data to the queryset.

        Individual columns can define additional joins and extra info to add on
        to the queryset. This handles adding all of those.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting augmented QuerySet.
        """
        for column in self.columns:
            try:
                queryset = column.augment_queryset(queryset)
            except Exception as e:
                logger.exception('Error when calling augment_queryset for '
                                 'DataGrid Column %r: %s',
                                 column, e)

        return queryset

    def render_listview(self, render_context=None):
        """Render the standard list view of the grid.

        This can be called from templates.

        Args:
            render_context (Context):
                The common template variable context to render on the datagrid,
                provided in the constructor.

        Returns:
            unicode: The rendered HTML for the datagrid page.
        """
        try:
            if render_context is None:
                render_context = self._build_render_context()

            self.load_state(render_context)

            context = {
                'datagrid': self,
            }

            context.update(self.extra_context)
            context.update(render_context)

            return render_to_string(self.listview_template, context)
        except Exception:
            trace = traceback.format_exc()
            logger.exception('Failed to render datagrid:\n%s',
                             trace,
                             request=self.request)
            return format_html('<pre>{0}</pre>', trace)

    def render_listview_to_response(self, request=None, render_context=None):
        """Render the listview to a response.

        The rendered result will not be cached by the browser.

        Args:
            request (HttpRequest):
                The HTTP request from the client.

            render_context (Context):
                The common template variable context to render on the datagrid,
                provided in the constructor.

        Returns:
            HttpResponse: The HTTP response to send to the client.
        """
        response = HttpResponse(
            str(self.render_listview(render_context)))
        patch_cache_control(response, no_cache=True, no_store=True, max_age=0,
                            must_revalidate=True)
        return response

    def render_to_response(self, template_name, extra_context={}):
        """Render the entire datagrid page to a response.

        This will render the entire page, given the specified template, with
        the datagrid as a part of it. This is the primary function a view
        will be using to render the page.

        Args:
            template_name (unicode):
                The template for the page.

            extra_context (dict):
                Extra context variables to use in the template.

        Returns:
            HttpResponse: The HTTP response to send to the client.
        """
        render_context = self._build_render_context()
        self.load_state(render_context)

        # If the caller is requesting just this particular grid, return it.
        if self.request.GET.get('gridonly', False) and \
           self.request.GET.get('datagrid-id', None) == self.id:
            return self.render_listview_to_response(
                render_context=render_context)

        context = {
            'datagrid': self
        }
        context.update(extra_context)
        context.update(render_context)

        return HttpResponse(render_to_string(template_name=template_name,
                                             context=context))

    def render_paginator(self, adjacent_pages=3):
        """Render the paginator for the datagrid.

        This can be called from templates.

        Args:
            adjacent_pages (int):
                The number of adjacent page numbers to show in the
                paginator.

        Returns:
            unicode: The paginator as HTML.
        """
        extra_query = get_url_params_except(self.request.GET,
                                            'page', 'gridonly',
                                            *self.special_query_args)

        page_nums = range(max(1, self.page.number - adjacent_pages),
                          min(self.paginator.num_pages,
                              self.page.number + adjacent_pages)
                          + 1)

        if extra_query:
            extra_query += '&'

        context = {
            'is_paginated': self.page.has_other_pages(),
            'hits': self.paginator.count,
            'results_per_page': self.paginate_by,
            'page': self.page.number,
            'pages': self.paginator.num_pages,
            'page_numbers': page_nums,
            'has_next': self.page.has_next(),
            'has_previous': self.page.has_previous(),
            'show_first': 1 not in page_nums,
            'show_last': self.paginator.num_pages not in page_nums,
            'extra_query': extra_query,
        }

        if self.page.has_next():
            context['next'] = self.page.next_page_number()
        else:
            context['next'] = None

        if self.page.has_previous():
            context['previous'] = self.page.previous_page_number()
        else:
            context['previous'] = None

        context.update(self.extra_context)

        return render_to_string(self.paginator_template, context)

    def build_paginator(self, queryset):
        """Build the paginator for the datagrid.

        This can be overridden to use a special paginator or to perform
        any kind of processing before passing on the query.

        Args:
            queryset (object):
                A queryset-compatible object.

        Returns:
            A populated paginator object.
        """
        return Paginator(queryset, self.paginate_by, self.paginate_orphans)

    def _build_render_context(self):
        """Build a dictionary containing RequestContext contents.

        A RequestContext can be expensive, so it's best to reuse the
        contents of one when possible. This is not easy with a standard
        RequestContext, but it's possible to build one and then pull out
        the contents into a dictionary.
        """
        request = self.request
        render_context = {}

        for context_processor in get_default_template_context_processors():
            render_context.update(context_processor(request))

        return render_context

    @staticmethod
    def link_to_object(state, obj, value):
        """Return a URL for the given object.

        This defaults to calling ``obj.get_absolute_url``.

        Returns:
            unicode: The URL for the object.
        """
        return obj.get_absolute_url()

    @staticmethod
    def link_to_value(state, obj, value):
        """Return a URL for the given value.

        This defaults to calling ``value.get_absolute_url``.

        Returns:
            unicode: The URL for the value.
        """
        return value.get_absolute_url()


class AlphanumericDataGrid(DataGrid):
    """A DataGrid subclass for an alphanumerically-paginated datagrid.

    This is useful for datasets that need to be queried alphanumerically,
    according to the starting character of their ``sortable`` column.
    """

    def __init__(self, request, queryset, sortable_column,
                 extra_regex='^[0-9].*', *args, **kwargs):
        """Initialize the datagrid.

        Args:
            request (HttpRequest):
                The HTTP request from the client.

            queryset (QuerySet):
                A QuerySet returning the objects to render in the grid.

            sortable_column (unicode):
                The model field used for the alphanumeric prefixes.

            extra_regex (unicode):
                A regex used for matching the beginning of entries in
                ``sortable_column``.
        """
        self.current_letter = request.GET.get('letter', 'all')

        regex_match = re.compile(extra_regex)

        if self.current_letter == 'all':
            pass  # No filtering
        elif self.current_letter.isalpha():
            queryset = queryset.filter(**{
                sortable_column + '__istartswith': self.current_letter
            })
        elif regex_match.match(self.current_letter):
            queryset = queryset.filter(**{
                sortable_column + '__regex': extra_regex
            })
        else:
            raise Http404

        super(AlphanumericDataGrid, self).__init__(request, queryset,
                                                   *args, **kwargs)

        self.extra_context['current_letter'] = self.current_letter
        self.extra_context['letters'] = (['all', '0'] +
                                         list(string.ascii_uppercase))

        self.special_query_args.append('letter')
        self.paginator_template = 'datagrid/alphanumeric_paginator.html'



from django.urls import include, path, re_path


def build_integration_urlpatterns(list_view_cls, config_form_view_cls,
                                  namespace=None):
    """Build URL patterns for integration pages.

    This will produce a set of URL patterns for the integration administration
    pages that take advantage of the specialized views provided. These can
    be used in a :file:`urls.py`, either returned as a ``urlpatterns``
    variable or included using :py:func:`~django.conf.urls.include`.

    Args:
        list_view_cls (type, optional):
            The view responsible for listing available integrations. This
            should be a subclass of
            :py:class:`~djblets.integrations.views.BaseIntegrationListView`.

            If ``None``, the integration list URL will not be provided in the
            URL patterns. Consumers should define their own
            ``integration-list`` URL with the appropriate namespace in order to
            handle any URL reversing that may occur.

        config_form_view_cls (type):
            The view responsible for configuring an integration. This should
            be a subclass of
            :py:class:`~djblets.integrations.views.
            BaseIntegrationConfigFormView`.

            If ``None``, the integration configuration URLs will not be
            provided in the URL patterns. This *will* break the integrations
            list URL unless consumers provide their own
            ``integration-add-config`` and ``integration-change-config`` URL
            registrations.

        namespace (unicode, optional):
            The namespace to use for all URL names.

            For legacy reasons, this is an empty string. If embedding into
            other parts of the app, this should be changed. For instance,
            you may want the admin version to use ``admin``.
    """
    urlpatterns = []

    if list_view_cls is not None:
        urlpatterns.append(
            path('', list_view_cls.as_view(), name='integration-list'))

    if config_form_view_cls is not None:
        urlpatterns.append(
            re_path(r'^(?P<integration_id>[A-Za-z0-9_\.]+)/configs/', include([
                path('add/',
                     config_form_view_cls.as_view(),
                     name='integration-add-config'),
                path('<int:config_id>/',
                     config_form_view_cls.as_view(),
                     name='integration-change-config'),
            ])))

    return [
        path('', include(urlpatterns, namespace=namespace)),
    ]



"""Unit tests for djblets.secrets.crypto."""

import inspect

from django.test.utils import override_settings

from djblets.secrets.crypto import (aes_decrypt,
                                    aes_decrypt_base64,
                                    aes_decrypt_iter,
                                    aes_encrypt,
                                    aes_encrypt_base64,
                                    aes_encrypt_iter,
                                    get_default_aes_encryption_key)
from djblets.testing.testcases import TestCase


@override_settings(SECRET_KEY='abcdefghijklmnopqrstuvwxyz012345')
class BaseAESTestCase(TestCase):
    """Base testcase for AES-related tests."""

    PLAIN_TEXT_UNICODE = 'this is a test 123 ^&*'
    PLAIN_TEXT_BYTES = b'this is a test 123 ^&*'

    PASSWORD_UNICODE = 'this is a test 123 ^&*'
    PASSWORD_BYTES = b'this is a test 123 ^&*'

    CUSTOM_KEY = b'0123456789abcdef'


class AESDecryptTests(BaseAESTestCase):
    """Unit tests for aes_decrypt."""

    def test_with_bytes(self):
        """Testing aes_decrypt with byte string"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = (
            b'\xfb\xdc\xb5h\x15\xa1\xb2\xdc\xec\xf1\x14\xa9\xc6\xab\xb2J\x10'
            b'\'\xd4\xf6&\xd4k9\x82\xf6\xb5\x8bmu\xc8E\x9c\xac\xc5\x04@B'
        )

        decrypted = aes_decrypt(encrypted)
        self.assertIsInstance(decrypted, bytes)
        self.assertEqual(decrypted, self.PLAIN_TEXT_BYTES)

    def test_with_unicode(self):
        """Testing aes_decrypt with Unicode string"""
        expected_message = ('The data to decrypt must be of type "bytes", '
                            'not "<class \'str\'>"')

        with self.assertRaisesMessage(TypeError, expected_message):
            aes_decrypt('abc')  # type: ignore

    def test_with_custom_key(self):
        """Testing aes_decrypt with custom key"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = (
            b'\x9cd$e\xb1\x9e\xe0z\xb8[\x9e!\xf2h\x90\x8d\x82f%G4\xc2\xf0'
            b'\xda\x8dr\x81ER?S6\x12%7\x98\x89\x90'
        )

        decrypted = aes_decrypt(encrypted, key=self.CUSTOM_KEY)
        self.assertIsInstance(decrypted, bytes)
        self.assertEqual(decrypted, self.PLAIN_TEXT_BYTES)

    def test_with_custom_key_unicode(self):
        """Testing aes_decrypt with custom key as Unicode"""
        encrypted = (
            b'\x9cd$e\xb1\x9e\xe0z\xb8[\x9e!\xf2h\x90\x8d\x82f%G4\xc2\xf0'
            b'\xda\x8dr\x81ER?S6\x12%7\x98\x89\x90'
        )
        expected_message = ('The encryption key must be of type "bytes", '
                            'not "<class \'str\'>"')

        with self.assertRaisesMessage(TypeError, expected_message):
            aes_decrypt(encrypted, key='abc')  # type: ignore


class AESDecryptBase64(BaseAESTestCase):
    """Unit tests for aes_decrypt_base64."""

    def test_with_bytes(self):
        """Testing aes_decrypt_base64 with byte string"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = b'AjsUGevO3UiVH7iN3zO9vxvqr5X5ozuAbOUByTATsitkhsih1Zc='
        decrypted = aes_decrypt_base64(encrypted)

        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)

    def test_with_unicode(self):
        """Testing aes_decrypt_base64 with Unicode string"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = 'AjsUGevO3UiVH7iN3zO9vxvqr5X5ozuAbOUByTATsitkhsih1Zc='
        decrypted = aes_decrypt_base64(encrypted)

        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)

    def test_with_custom_key(self):
        """Testing aes_decrypt_base64 with custom key"""
        # The encrypted value was made with PyCrypto, to help with
        # compatibility testing from older installs.
        encrypted = b'/pOO3VWHRXd1ZAeHZo8MBGQsNClD4lS7XK9WAydt8zW/ob+e63E='
        decrypted = aes_decrypt_base64(encrypted, key=self.CUSTOM_KEY)

        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)


class AESDecryptIterTests(BaseAESTestCase):
    """Unit tests for aes_decrypt_iter."""

    def test_with_bytes(self):
        """Testing aes_decrypt_iter with byte string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        def _gen_data():
            yield (b'\xfd\x1c\xccF\xe8\xa8\xe3\x94`\xfa\xcf\xc7\x11\xeabD'
                   b'\xdft\xfa;Y\x06\xb2\xb3\xae\x10#"kR\x13')
            yield b'\xc7\x17/\x02w\xe0\xe4\xa32L/\x9d\r\xd1v\xa2\xec\xb41\x82'
            yield (b'\xa8\x04\xa9M\t\xac\x92\x9d|\xc0\xb3\xa4\x1f+\xab\x0c\t'
                   b'\xc4\x80\x8c')
            yield b'*'

        decrypted = aes_decrypt_iter(_gen_data())

        self.assertTrue(inspect.isgenerator(decrypted))
        self.assertEqual(b''.join(decrypted),
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')

    def test_with_list(self):
        """Testing aes_decrypt_iter with list"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        decrypted = aes_decrypt_iter([
            (b'\xfd\x1c\xccF\xe8\xa8\xe3\x94`\xfa\xcf\xc7\x11\xeabD'
             b'\xdft\xfa;Y\x06\xb2\xb3\xae\x10#"kR\x13'),

            b'\xc7\x17/\x02w\xe0\xe4\xa32L/\x9d\r\xd1v\xa2\xec\xb41\x82',

            (b'\xa8\x04\xa9M\t\xac\x92\x9d|\xc0\xb3\xa4\x1f+\xab\x0c\t'
             b'\xc4\x80\x8c'),

            b'*',
        ])

        self.assertTrue(inspect.isgenerator(decrypted))
        self.assertEqual(b''.join(decrypted),
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')

    def test_with_empty_iter(self):
        """Testing aes_encrypt_iter with empty iterator"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        decrypted = aes_decrypt_iter([])

        self.assertTrue(inspect.isgenerator(decrypted))

        with self.assertRaisesMessage(ValueError,
                                      'Invalid IV size (0) for CFB8.'):
            list(decrypted)

    def test_with_empty_str(self):
        """Testing aes_encrypt_iter with only empty string from iterator"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        decrypted = aes_decrypt_iter([b''])

        self.assertTrue(inspect.isgenerator(decrypted))

        with self.assertRaisesMessage(ValueError,
                                      'Invalid IV size (0) for CFB8.'):
            list(decrypted)

    def test_with_iv_data_split(self):
        """Testing aes_encrypt_iter with IV and encrypted data split across
        items
        """
        def _gen_data():
            data = (
                b'\xfd\x1c\xccF\xe8\xa8\xe3\x94`\xfa\xcf\xc7\x11\xeabD\xdft'
                b'\xfa;Y\x06\xb2\xb3\xae\x10#"kR\x13\xc7\x17/\x02w\xe0\xe4'
                b'\xa32L/\x9d\r\xd1v\xa2\xec\xb41\x82\xa8\x04\xa9M\t\xac\x92'
                b'\x9d|\xc0\xb3\xa4\x1f+\xab\x0c\t\xc4\x80\x8c*'
            )

            # Iterating over a byte string yields integers, unless we slice.
            # Hence needing to do this.
            yield from (
                data[i:i + 1]
                for i in range(len(data))
            )

        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        decrypted = aes_decrypt_iter(_gen_data())

        self.assertTrue(inspect.isgenerator(decrypted))
        self.assertEqual(b''.join(decrypted),
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')

    def test_with_just_iv_finalizer(self):
        """Testing aes_encrypt_iter with only IV + finalizer in encypted data
        """
        def _gen_data():
            yield b'\xa9;\xb1)b\xaeSF.S\xb4\x8f\x8d\xecT\xfb'

        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        decrypted = aes_decrypt_iter(_gen_data())

        self.assertTrue(inspect.isgenerator(decrypted))
        self.assertEqual(b''.join(decrypted),
                         b'')

    def test_with_custom_key(self):
        """Testing aes_decrypt_iter with custom key"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        def _gen_data():
            yield (b'\xbbq\xb8w\xf8s!\x10E4+\x89^0\x00\x04A\xcc\x82"\xa7\xc6'
                   b'\xe0\xfe\x93M\x0b\xdc\xdf\xe1\xfb')
            yield (b'?\x06\xef\x94\xda\x88\xb7\xd0\xbfM\xc8@\x80\xff\xde\x8a'
                   b'\xc4\xab\xb9h')
            yield (b'B\x97\x82am\x9b\x8d\xf0\x8c\x10\xfa\xb1\xf7\xc7\x9c\x97'
                   b'9ZbA')
            yield b'\x01'

        decrypted = aes_decrypt_iter(_gen_data(),
                                     key=self.CUSTOM_KEY)

        self.assertTrue(inspect.isgenerator(decrypted))
        self.assertEqual(b''.join(decrypted),
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')


class AESEncryptTests(BaseAESTestCase):
    """Unit tests for aes_encrypt."""

    def test_with_bytes(self):
        """Testing aes_encrypt with byte string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt(self.PLAIN_TEXT_BYTES)
        self.assertIsInstance(encrypted, bytes)
        self.assertEqual(aes_decrypt(encrypted), self.PLAIN_TEXT_BYTES)

    def test_with_unicode(self):
        """Testing aes_encrypt with Unicode string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt(self.PLAIN_TEXT_UNICODE)
        self.assertIsInstance(encrypted, bytes)
        self.assertEqual(aes_decrypt(encrypted), self.PLAIN_TEXT_BYTES)

    def test_with_custom_key(self):
        """Testing aes_encrypt with custom key"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt(self.PLAIN_TEXT_BYTES, key=self.CUSTOM_KEY)
        decrypted = aes_decrypt(encrypted, key=self.CUSTOM_KEY)

        self.assertIsInstance(decrypted, bytes)
        self.assertEqual(decrypted, self.PLAIN_TEXT_BYTES)


class AESEncryptBase64(BaseAESTestCase):
    """Unit tests for aes_encrypt_base64."""

    def test_with_bytes(self):
        """Testing aes_encrypt_base64 with byte string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_decrypt_base64(
            aes_encrypt_base64(self.PASSWORD_BYTES))
        self.assertIsInstance(encrypted, str)
        self.assertEqual(encrypted, self.PASSWORD_UNICODE)

    def test_with_unicode(self):
        """Testing aes_encrypt_base64 with Unicode string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_decrypt_base64(
            aes_encrypt_base64(self.PASSWORD_UNICODE))
        self.assertIsInstance(encrypted, str)
        self.assertEqual(encrypted, self.PASSWORD_UNICODE)

    def test_with_custom_key(self):
        """Testing aes_encrypt_base64 with custom key"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt_base64(self.PASSWORD_UNICODE,
                                       key=self.CUSTOM_KEY)
        self.assertIsInstance(encrypted, str)

        decrypted = aes_decrypt_base64(encrypted, key=self.CUSTOM_KEY)
        self.assertIsInstance(decrypted, str)
        self.assertEqual(decrypted, self.PASSWORD_UNICODE)


class AESEncryptIterTests(BaseAESTestCase):
    """Unit tests for aes_encrypt_iter."""

    def test_with_bytes(self):
        """Testing aes_encrypt_iter with byte string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        def _gen_data():
            yield b'this is a test\n'
            yield b'and this is another\n'
            yield b'hey look three tests'
            yield b'!'

        encrypted = aes_encrypt_iter(_gen_data())

        self.assertTrue(inspect.isgenerator(encrypted))
        self.assertEqual(aes_decrypt(b''.join(encrypted)),
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')

    def test_with_unicode(self):
        """Testing aes_encrypt_iter with Unicode string"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        def _gen_data():
            yield 'this is a test\n'
            yield 'and this is another\n'
            yield 'hey look three tests'
            yield '!'

        encrypted = aes_encrypt_iter(_gen_data())

        self.assertTrue(inspect.isgenerator(encrypted))
        self.assertEqual(aes_decrypt(b''.join(encrypted)),
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')

    def test_with_list(self):
        """Testing aes_encrypt_iter with list"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        encrypted = aes_encrypt_iter([
            b'this is a test\n',
            b'and this is another\n',
            b'hey look three tests',
            b'!',
        ])

        self.assertTrue(inspect.isgenerator(encrypted))
        self.assertEqual(aes_decrypt(b''.join(encrypted)),
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')

    def test_with_empty_iter(self):
        """Testing aes_encrypt_iter with empty iterator"""
        encrypted = aes_encrypt_iter([])

        self.assertTrue(inspect.isgenerator(encrypted))
        self.assertEqual(aes_decrypt(b''.join(encrypted)),
                         b'')

    def test_with_custom_key(self):
        """Testing aes_encrypt_iter with custom key"""
        # The encrypted value will change every time, since the iv changes,
        # so we can't compare a direct value. Instead, we need to ensure that
        # we can decrypt what we encrypt.
        def _gen_data():
            yield b'this is a test\n'
            yield b'and this is another\n'
            yield b'hey look three tests'
            yield b'!'

        encrypted = aes_encrypt_iter(_gen_data(), key=self.CUSTOM_KEY)
        decrypted = aes_decrypt(b''.join(encrypted), key=self.CUSTOM_KEY)

        self.assertIsInstance(decrypted, bytes)
        self.assertEqual(decrypted,
                         b'this is a test\n'
                         b'and this is another\n'
                         b'hey look three tests!')


class GetDefaultAESEncryptionKeyTests(BaseAESTestCase):
    """Unit tests for get_default_aes_encryption_key."""

    def test_get_default_aes_encryption_key(self):
        """Testing get_default_aes_encryption_key"""
        key = get_default_aes_encryption_key()
        self.assertIsInstance(key, bytes)
        self.assertEqual(key, b'abcdefghijklmnop')



import io
import re
import sys
from html.entities import name2codepoint
from xml.dom.minidom import parseString

from markdown import markdownFromFile


MARKDOWN_ALL_ESCAPED_CHARS = (
    '\\', '`', '*', '_', '{', '}', '[', ']', '(', ')', '>', '#',
    '+', '-', '.', '!',
)

MARKDOWN_SPECIAL_CHARS = re.escape(r''.join(MARKDOWN_ALL_ESCAPED_CHARS))
MARKDOWN_SPECIAL_CHARS_RE = re.compile(r'([%s])' % MARKDOWN_SPECIAL_CHARS)

# Markdown.ESCAPED_CHARS lists several characters to escape, but it's not
# that simple. We only want to escape certain things if they'll actually affect
# the Markdown rendering, because otherwise it's annoying to look at the
# source.
MARKDOWN_ESCAPED_CHARS = set(MARKDOWN_ALL_ESCAPED_CHARS) - {
    '.', '#', '-', '+', '_', '(', ')', '*', '>'
}

ESCAPE_CHARS_RE = re.compile(r"""
    (
    # Numeric lists start with leading whitespace, one or more digits,
    # and then a period
      ^\s*(\d+\.)\s

    # ATX-style headers start with a hash at the beginning of the line.
    | ^\s*(\#+)

    # + and - have special meaning (lists, headers, and rules), but only if
    # they're at the start of the line.
    | ^\s*([-\+]+)

    # _ indicates italic, and __ indicates bold, but not when in the middle
    # of a word.
    #
    | (?<!\w|_)(__?)
    | (__?)(?!\w|_)

    # This is an alternate format for italic and bold, using * instead of _.
    | (?<!\w|\*)(\*\*?)
    | (\*\*?)(?!\w|\*)

    # Named links are in the form of [name](url).
    | (\[) [^\]]* (\]) (\() [^\)]* (\))

    # '>' need only be escaped for blockquotes ('> ...') or automatic links
    # ('<http://...> or <user@example.com>).
    | ^((?:\s*>)+)
    | (?:<(?:(?:[Ff]|[Hh][Tt])[Tt][Pp][Ss]?://[^>]*))(>)
    | (?:<[^> \!]*@[^> ]*)(>)

    # All other special characters
    | ([%s])
    )
    """ % re.escape(''.join(MARKDOWN_ESCAPED_CHARS)),
    re.M | re.VERBOSE)
UNESCAPE_CHARS_RE = re.compile(r'\\([%s])' % MARKDOWN_SPECIAL_CHARS)

ILLEGAL_XML_CHARS_RE = None
_ILLEGAL_XML_CHAR_CODES_SET = None


def markdown_escape(text):
    """Escape text for use in Markdown.

    This will escape the provided text so that none of the characters will
    be rendered specially by Markdown.
    """
    def _escape_matches(m):
        prev_end = m.start(0)
        new_s = []

        for i, group in enumerate(m.groups()[1:], start=2):
            if group:
                new_s.append(m.string[prev_end:m.start(i)])
                new_s.append(MARKDOWN_SPECIAL_CHARS_RE.sub(r'\\\1', group))
                prev_end = m.end(i)

        new_s.append(m.string[prev_end:m.end(0)])

        return ''.join(new_s)

    return ESCAPE_CHARS_RE.sub(_escape_matches, text)


def markdown_unescape(escaped_text):
    """Unescape Markdown-escaped text.

    This will unescape the provided Markdown-formatted text so that any
    escaped characters will be unescaped.
    """
    text = UNESCAPE_CHARS_RE.sub(r'\1', escaped_text)

    split = text.split('\n')
    for i, line in enumerate(split):
        if line.startswith('&nbsp;   '):
            split[i] = ' ' + line[6:]
        elif line.startswith('&nbsp;\t'):
            split[i] = line[6:]

    return '\n'.join(split)


def iter_markdown_lines(markdown_html):
    """Iterate over lines of Markdown, normalizing for individual display.

    Generated Markdown HTML cannot by itself be handled on a per-line-basis.
    Code blocks, for example, will consist of multiple lines of content
    contained within a <pre> tag. Likewise, lists will be a bunch of
    <li> tags inside a <ul> tag, and individually do not form valid lists.

    This function iterates through the Markdown tree and generates
    self-contained lines of HTML that can be rendered individually.
    """
    nodes = get_markdown_element_tree(markdown_html)

    for node in nodes:
        if node.nodeType == node.ELEMENT_NODE:
            if (node.tagName == 'div' and
                node.attributes.get('class', 'codehilite')):
                # This is a code block, which will consist of a bunch of lines
                # for the source code. It contains a single <pre> element
                # within it that contains the full code block. We want to split
                # that up into individual lines which are each wrapped in their
                # own <div class="codehilite"><pre> elements.
                child_lines = node.firstChild.toxml().splitlines()

                # The first and last lines in child_lines begin and end with
                # <pre> and </pre>, respectively. Clear that out.
                assert child_lines[0].startswith('<pre>')
                assert child_lines[-1].endswith('</pre>')
                child_lines[0] = child_lines[0][len('<pre>'):]
                child_lines[-1] = child_lines[-1][:-len('</pre>')]
                if not child_lines[-1]:
                    child_lines = child_lines[:-1]

                fmt = '<div class="codehilite%s"><pre>%s</pre></div>'

                # If there's just a single line, output that directly
                if len(child_lines) == 1:
                    yield fmt % ('', child_lines[0])
                elif len(child_lines) > 1:
                    yield fmt % (' codehilite-multiline-start', child_lines[0])

                    for line in child_lines[1:-1]:
                        yield fmt % (' codehilite-multiline-middle', line)

                    yield fmt % (' codehilite-multiline-end', child_lines[-1])
            elif node.tagName in ('ul', 'ol'):
                # This is a list. We'll need to split all of its items
                # into individual lists, in order to retain bullet points
                # or the numbers.
                #
                # For the case of numbers, we can set each list to start
                # at the appropriate number so that they don't all say "1."
                start = node.attributes.get('start')
                if start is not None:
                    try:
                        i = int(start.value)
                    except ValueError:
                        i = 1
                else:
                    i = 1

                for child_node in node.childNodes:
                    if (child_node.nodeType == child_node.ELEMENT_NODE and
                        child_node.tagName == 'li'):
                        # This is a list item element. It may be multiple
                        # lines, but we'll have to treat it as one line.
                        yield '<%s start="%s">%s</%s>' % (
                            node.tagName, i, child_node.toxml(),
                            node.tagName)

                        i += 1
            elif node.tagName == 'p':
                # This is a paragraph, possibly containing multiple lines.
                for line in node.toxml().splitlines():
                    yield line
            else:
                # Whatever this is, treat it as one block.
                yield node.toxml()
        elif node.nodeType == node.TEXT_NODE:
            # This may be several blank extraneous blank lines, due to
            # Markdown's generation from invisible markup like fences.
            # We want to condense this down to one blank line.
            yield '\n'


def get_markdown_element_tree(markdown_html):
    """Return an XML element tree for Markdown-generated HTML.

    This will build the tree and return all nodes representing the rendered
    Markdown content. The element tree is generated using Python's
    :py:mod:`xml.dom.minidom`.

    Args:
        markdown_html (bytes or unicode):
            The Markdown-generated HTML to parse.

    Returns:
        xml.dom.minicompat.NodeList:
            The list of nodes representing the Markdown-generated HTML.
    """
    markdown_html = sanitize_illegal_chars_for_xml(markdown_html)

    # Python-Markdown converts all characters in an e-mail address to
    # entities (likely to avoid e-mail address harvesting). It optimistically
    # tries to use named HTML entities for special characters, but these won't
    # be known to an XML parser.
    #
    # Since there's no easy way to register entities with xml.dom.minidom, and
    # we don't want to change the return type, the easiest solution is to
    # convert named entites back to character codes through a regex before
    # parsing.
    illegal_chars_set = _get_illegal_char_codes_set_for_xml()
    unknown_code = ord('?')
    unknown_entity = '#%s' % unknown_code

    def _handle_entity(m):
        entity = m.group(1)

        if entity.startswith('#'):
            try:
                if entity.startswith('#x'):
                    value = int(entity[2:], 16)
                else:
                    value = int(entity[1:])

                if value in illegal_chars_set:
                    entity = unknown_entity
            except ValueError:
                entity = unknown_entity

            return '&%s;' % entity
        else:
            return '&#%s;' % name2codepoint.get(entity, unknown_code)

    markdown_html = re.sub(r'&([^;]+);', _handle_entity, markdown_html,
                           flags=re.M)

    doc = parseString(('<html>%s</html>' % markdown_html).encode('utf-8'))

    return doc.childNodes[0].childNodes


def sanitize_illegal_chars_for_xml(s):
    """Sanitize a string, removing characters illegal in XML.

    This will remove a number of characters that would break the XML parser.
    They may be in the string due to a copy/paste.

    Note that this will not perform any security-related sanitization of the
    HTML. It's purely a parsing aid for dealing with illegal characters.

    This code is courtesy of the XmlRpcPlugin developers, as documented
    here: http://stackoverflow.com/a/22273639

    Args:
        s (unicode):
            The string to sanitize.

    Returns:
        unicode:
        The resulting sanitized HTML.
    """
    global ILLEGAL_XML_CHARS_RE

    if ILLEGAL_XML_CHARS_RE is None:
        ILLEGAL_XML_CHARS_RE = re.compile('[%s]' % ''.join(
            '%s-%s' % (chr(low), chr(high))
            for low, high in _get_illegal_chars_for_xml()
        ))

    if isinstance(s, bytes):
        s = s.decode('utf-8')

    return ILLEGAL_XML_CHARS_RE.sub('', s)


def render_markdown_from_file(f, **markdown_kwargs):
    """Render Markdown text from a file stream to HTML.

    Args:
        f (file or io.BytesIO):
            The byte stream to read from.

        **markdown_kwargs (dict):
            Keyword arguments to pass to :py:func:`markdown.markdownFromFile`.

    Returns:
        unicode:
        The resulting Markdown-rendered HTML.
    """
    s = io.BytesIO()

    try:
        markdownFromFile(input=f, output=s, **markdown_kwargs)

        return s.getvalue().decode('utf-8')
    finally:
        s.close()


def _get_illegal_char_codes_set_for_xml():
    """Return a set of all illegal character codes for an XML document.

    The set will contain a numeric character code for every single character
    that's considered invalid in an XML document.

    Version Added:
        1.0.18

    Returns:
        set of int:
        A set of all illegal character codes.
    """
    global _ILLEGAL_XML_CHAR_CODES_SET

    if _ILLEGAL_XML_CHAR_CODES_SET is None:
        _ILLEGAL_XML_CHAR_CODES_SET = set()

        for illegal_range in _get_illegal_chars_for_xml():
            _ILLEGAL_XML_CHAR_CODES_SET.update(range(illegal_range[0],
                                                     illegal_range[1] + 1))

    return _ILLEGAL_XML_CHAR_CODES_SET


def _get_illegal_chars_for_xml():
    """Return all illegal character ranges for an XML document.

    This is used internally to build ranges of characters that aren't
    valid in an XML document, for the purposes of building regexes and
    maps for filtering them out.

    We don't cache the result, because this will only ever be generated
    up to two times per process. The end results themselves are cached, and
    there's no point in keeping the raw lists in memory.

    Version Added:
        1.0.18

    Returns:
        list of tuple:
        The list of inclusive illegal character ranges.
    """
    illegal_unichrs = [
        (0x00, 0x08), (0x0B, 0x0C), (0x0E, 0x1F), (0x7F, 0x84),
        (0x86, 0x9F), (0xFDD0, 0xFDDF), (0xFFFE, 0xFFFF),
    ]

    if sys.maxunicode > 0x10000:
        illegal_unichrs += [
            (0x1FFFE, 0x1FFFF), (0x2FFFE, 0x2FFFF), (0x3FFFE, 0x3FFFF),
            (0x4FFFE, 0x4FFFF), (0x5FFFE, 0x5FFFF), (0x6FFFE, 0x6FFFF),
            (0x7FFFE, 0x7FFFF), (0x8FFFE, 0x8FFFF), (0x9FFFE, 0x9FFFF),
            (0xAFFFE, 0xAFFFF), (0xBFFFE, 0xBFFFF), (0xCFFFE, 0xCFFFF),
            (0xDFFFE, 0xDFFFF), (0xEFFFE, 0xEFFFF), (0xFFFFE, 0xFFFFF),
            (0x10FFFE, 0x10FFFF),
        ]

    return illegal_unichrs



"""Views for the log app."""

import calendar
import codecs
import datetime
import logging
import os
import re
import time
from urllib.parse import urlencode

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404
from django.shortcuts import render
from django.utils.translation import gettext_lazy as _


LEVELS = (
    (logging.DEBUG, 'debug', _('Debug')),
    (logging.INFO, 'info', _('Info')),
    (logging.WARNING, 'warning', _('Warning')),
    (logging.ERROR, 'error', _('Error')),
    (logging.CRITICAL, 'critical', _('Critical')),
)


# Matches the default timestamp format in the logging module.
TIMESTAMP_FMT = '%Y-%m-%d %H:%M:%S'

LOG_LINE_RE = re.compile(
    r'(?P<timestamp>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3}) - '
    r'(?P<level>DEBUG|INFO|WARNING|ERROR|CRITICAL) - '
    r'(?P<message>.*)')


def parse_timestamp(format, timestamp_str):
    """Utility function to parse a timestamp into a datetime.datetime.

    Python 2.5 and up have datetime.strptime, but Python 2.4 does not,
    so we roll our own as per the documentation.

    If passed a timestamp_str of None, we will return None as a convenience.
    """
    if not timestamp_str:
        return None

    return datetime.datetime(*time.strptime(timestamp_str, format)[0:6])


def build_query_string(request, params):
    """Builds a query string that includes the specified parameters along
    with those that were passed to the page.

    params is a dictionary.
    """
    query_parts = []

    for key, value in request.GET.items():
        if key not in params:
            query_parts.append(urlencode({
                key: value
            }))

    for key, value in params.items():
        if value is not None:
            query_parts.append(urlencode({
                key: value
            }))

    return '?' + '&'.join(query_parts)


def iter_log_lines(from_timestamp, to_timestamp, requested_levels):
    """Generator that iterates over lines in a log file, yielding the
    yielding information about the lines."""
    log_filename = os.path.join(settings.LOGGING_DIRECTORY,
                                settings.LOGGING_NAME + '.log')

    line_info = None

    try:
        fp = codecs.open(log_filename, encoding='utf-8')
    except IOError:
        # We'd log this, but it'd do very little good in practice.
        # It would only appear on the console when using the development
        # server, but production users would never see anything. So,
        # just return gracefully. We'll show an empty log, which is
        # about accurate.
        return

    for line in fp:
        line = line.rstrip()

        m = LOG_LINE_RE.match(line)

        if m:
            if line_info:
                # We have a fully-formed log line and this new line isn't
                # part of it, so yield it now.
                yield line_info
                line_info = None

            timestamp_str = m.group('timestamp')
            level = m.group('level')
            message = m.group('message')

            if not requested_levels or level.lower() in requested_levels:
                timestamp = parse_timestamp(TIMESTAMP_FMT,
                                            timestamp_str.split(',')[0])

                timestamp_date = timestamp.date()

                if ((from_timestamp and from_timestamp > timestamp_date) or
                    (to_timestamp and to_timestamp < timestamp_date)):
                    continue

                line_info = (timestamp, level, message)
        elif line_info:
            line_info = (line_info[0],
                         line_info[1],
                         line_info[2] + "\n" + line)

    if line_info:
        yield line_info

    fp.close()


def get_log_filtersets(request, requested_levels,
                       from_timestamp, to_timestamp):
    """Returns the filtersets that will be used in the log view."""
    logger = logging.getLogger('')
    level_filters = [
        {
            'name': _('All'),
            'url': build_query_string(request, {'levels': None}),
            'selected': len(requested_levels) == 0,
        }
    ] + [
        {
            'name': label_name,
            'url': build_query_string(request, {'levels': level_name}),
            'selected': level_name in requested_levels,
        }
        for level_id, level_name, label_name in LEVELS
        if logger.isEnabledFor(level_id)
    ]

    from_timestamp_str = request.GET.get('from', None)
    to_timestamp_str = request.GET.get('to', None)
    today = datetime.date.today()
    today_str = today.strftime('%Y-%m-%d')
    one_week_ago = today - datetime.timedelta(days=7)
    one_week_ago_str = one_week_ago.strftime('%Y-%m-%d')
    month_range = calendar.monthrange(today.year, today.month)
    this_month_begin_str = today.strftime('%Y-%m-01')
    this_month_end_str = today.strftime('%Y-%m-') + str(month_range[1])

    date_filters = [
        {
            'name': _('Any date'),
            'url': build_query_string(request, {
                'from': None,
                'to': None,
            }),
            'selected': (from_timestamp_str is None and
                         to_timestamp_str is None),
        },
        {
            'name': _('Today'),
            'url': build_query_string(request, {
                'from': today_str,
                'to': today_str,
            }),
            'selected': (from_timestamp_str == today_str and
                         to_timestamp_str == today_str),
        },
        {
            'name': _('Past 7 days'),
            'url': build_query_string(request, {
                'from': one_week_ago_str,
                'to': today_str,
            }),
            'selected': (from_timestamp_str == one_week_ago_str and
                         to_timestamp_str == today_str),
        },
        {
            'name': _('This month'),
            'url': build_query_string(request, {
                'from': this_month_begin_str,
                'to': this_month_end_str,
            }),
            'selected': (from_timestamp_str == this_month_begin_str and
                         to_timestamp_str == this_month_end_str),
        },
    ]

    return (
        (_("By date"), date_filters),
        (_("By level"), level_filters),
    )


@staff_member_required
def server_log(request, template_name='log/log.html'):
    """Displays the server log."""

    # First check if logging is even configured. If it's not, just return
    # a 404.
    if (not getattr(settings, "LOGGING_ENABLED", False) or
        not getattr(settings, "LOGGING_DIRECTORY", None)):
        raise Http404()

    requested_levels = []

    # Get the list of levels to show.
    if 'levels' in request.GET:
        requested_levels = request.GET.get('levels').split(',')

    # Get the timestamp ranges.
    from_timestamp = parse_timestamp('%Y-%m-%d', request.GET.get('from'))
    to_timestamp = parse_timestamp('%Y-%m-%d', request.GET.get('to'))

    if from_timestamp:
        from_timestamp = from_timestamp.date()

    if to_timestamp:
        to_timestamp = to_timestamp.date()

    # Get the filters to show.
    filtersets = get_log_filtersets(request, requested_levels,
                                    from_timestamp, to_timestamp)

    # Grab the lines from the log file.
    log_lines = iter_log_lines(from_timestamp, to_timestamp, requested_levels)

    # Figure out the sorting
    sort_type = request.GET.get('sort', 'asc')

    if sort_type == 'asc':
        reverse_sort_type = 'desc'
    else:
        reverse_sort_type = 'asc'
        log_lines = reversed(list(log_lines))

    return render(request, template_name, {
        'log_lines': log_lines,
        'filtersets': filtersets,
        'sort_url': build_query_string(request, {'sort': reverse_sort_type}),
        'sort_type': sort_type,
    })



"""Conditions and sets of conditions."""

import logging

from django.utils.translation import gettext as _

from djblets.conditions.errors import (ConditionChoiceNotFoundError,
                                       ConditionOperatorNotFoundError,
                                       InvalidConditionModeError,
                                       InvalidConditionValueError)


class Condition(object):
    """A condition used to match state to a choice, operator, and value.

    Conditions store a choice, operator, and value (depending on the operator).
    Callers can query whether a value fulfills a given condition, making it
    easy for users to compose sets of rules safely for controlling behavior
    in an application without having to write any code.

    Generally, queries will be made against a :py:class:`ConditionSet`, instead
    of an individual Condition.

    Attributes:
        choice (djblets.conditions.choices.BaseConditionChoice):
            The choice stored for this condition.

        operator (djblets.conditions.operators.BaseConditionOperator):
            The operator stored for this condition.

        value (object):
            The value stored for this condition.

        raw_value (object):
            The raw (serialized) value for this condition. This is used
            internally, and won't usually be needed by a caller.
    """

    @classmethod
    def deserialize(cls, choices, data, condition_index=None,
                    choice_kwargs={}):
        """Deserialize a condition from serialized data.

        This expects data serialized by :py:meth:`serialize`.

        Args:
            choices (djblets.conditions.choices.ConditionChoices):
                Possible choices for the condition.

            data (dict):
                Serialized data representing this condition.

            condition_index (int, optional):
                The index of the condition within the set of conditions.
                This is used for exceptions to help identify which condition
                failed during deserialization.

        Returns:
            djblets.conditions.conditions.Condition:
            The deserialized condition.

        Raises:
            djblets.conditions.errors.ConditionChoiceNotFoundError:
                The choice ID referenced in the data was missing or did not
                match a valid choice.

            djblets.conditions.errors.ConditionOperatorNotFoundError:
                The operator ID referenced in the data was missing or did not
                match a valid operator for the choice.

            djblets.conditions.errors.InvalidConditionValueError:
                The value was missing from the payload data or was not valid
                for the choice and operator.
        """
        # Sanity-check that we have the data we expect.
        try:
            choice_id = data['choice']
        except KeyError:
            logging.debug('Condition.deserialize: Missing "choice" key for '
                          'condition %r',
                          data)

            raise ConditionChoiceNotFoundError(
                _('A choice is required.'),
                condition_index=condition_index)

        try:
            operator_id = data['op']
        except KeyError:
            logging.debug('Condition.deserialize: Missing "op" key for '
                          'condition %r',
                          data)

            raise ConditionOperatorNotFoundError(
                _('An operator is required.'),
                condition_index=condition_index)

        # Load the choice.
        try:
            choice = choices.get_choice(choice_id, choice_kwargs=choice_kwargs)
        except ConditionChoiceNotFoundError as e:
            logging.debug('Condition.deserialize: Invalid "choice" value '
                          '"%s" for condition %r',
                          choice_id, data)

            raise ConditionChoiceNotFoundError(
                str(e),
                choice_id=choice_id,
                condition_index=condition_index)

        # Load the operator.
        try:
            operator = choice.get_operator(operator_id)
        except ConditionOperatorNotFoundError as e:
            logging.debug('Condition.deserialize: Invalid "op" value "%s" '
                          'for condition %r',
                          operator_id, data)

            raise ConditionOperatorNotFoundError(
                str(e),
                operator_id=operator_id,
                condition_index=condition_index)

        # Load the value.
        if operator.value_field is not None:
            try:
                raw_value = data['value']
                value = operator.value_field.deserialize_value(raw_value)
            except KeyError:
                logging.debug('Condition.deserialize: Missing "value" value '
                              'for condition %r',
                              data)

                raise InvalidConditionValueError(
                    _('A value is required.'),
                    condition_index=condition_index)
            except InvalidConditionValueError as e:
                logging.debug('Condition.deserialize: Invalid "value" value '
                              '%r for condition %r',
                              raw_value, data)

                e.condition_index = condition_index

                raise
        else:
            raw_value = None
            value = None

        return cls(choice=choice,
                   operator=operator,
                   value=value,
                   raw_value=raw_value)

    def __init__(self, choice, operator, value=None, raw_value=None):
        """Initialize the condition.

        Args:
            choice (djblets.conditions.choices.BaseConditionChoice):
                The choice for this condition.

            operator (djblets.conditions.operators.BaseConditionOperator):
                The operator for this condition.

            value (object, optional):
                The value for this condition.

            raw_value (object, optional):
                The raw (serialized) value for this condition.
        """
        self.choice = choice
        self.operator = operator
        self.value = value

        if raw_value is None:
            self.raw_value = value
        else:
            self.raw_value = raw_value

    def matches(self, value, value_state_cache=None):
        """Return whether a value matches the condition.

        Args:
            value (object):
                The value to match against.

            value_state_cache (dict):
                An optional dictionary used to cache common computable data
                that might be shared across instances of one or more
                conditions.

        Returns:
            bool:
            ``True`` if the value fulfills the condition. ``False`` if it
            does not.
        """
        if value_state_cache is None:
            value_state_cache = {}

        return self.choice.matches(operator=self.operator,
                                   match_value=value,
                                   condition_value=self.value,
                                   value_state_cache=value_state_cache)

    def serialize(self):
        """Serialize the condition to a JSON-serializable dictionary.

        Returns:
            dict:
            A dictionary representing the condition. It can be safely
            serialized to JSON.
        """
        data = {
            'choice': self.choice.choice_id,
            'op': self.operator.operator_id,
        }

        if self.operator.value_field is not None:
            if self.value is None:
                value = None
            else:
                value = self.operator.value_field.serialize_value(self.value)

            data['value'] = value

        return data

    # Make this serializable in a DjbletsJSONEncoder.
    to_json = serialize


class ConditionSet(object):
    """A set of conditions used to match state and define rules.

    Condition sets own multiple conditions, and are given a mode indicating
    how to query state against those conditions. They're also responsible
    for serializing and deserializing all data around a set of conditions to
    a JSON-serializable format.

    If using :py:attr:`MODE_ALL`, then all conditions must be satisfied for a
    condition set to pass.  If using :py:attr:`MODE_ANY`, then only one
    condition must be satisfied.

    Attributes:
        mode (unicode):
            The matching mode for the condition set. This is one of
            :py:attr:`MODE_ALL` or :py:attr:`MODE_ANY`.

        conditions (list of Condition):
            The list of conditions that comprise this set.
    """

    #: Always match without conditions.
    MODE_ALWAYS = 'always'

    #: All conditions must match a value to satisfy the condition set.
    MODE_ALL = 'all'

    #: Any condition may match a value to satisfy the condition set.
    MODE_ANY = 'any'

    #: A set of all the valid modes.
    CONDITIONS = (MODE_ALWAYS, MODE_ALL, MODE_ANY)

    #: The default mode.
    DEFAULT_MODE = MODE_ALL

    @classmethod
    def deserialize(cls, choices, data, choice_kwargs={}):
        """Deserialize a set of conditions from serialized data.

        This expects data serialized by :py:meth:`deserialize`.

        Args:
            choices (djblets.conditions.choices.ConditionChoices):
                Possible choices for the condition set.

            data (dict):
                Serialized data representing this condition set.

        Returns:
            djblets.conditions.conditions.ConditionSet:
            The deserialized condition set.

        Raises:
            djblets.conditions.errors.ConditionChoiceNotFoundError:
                The choice ID referenced in the data was missing or did not
                match a valid choice in a condition.

            djblets.conditions.errors.ConditionOperatorNotFoundError:
                The operator ID referenced in the data was missing or did not
                match a valid operator for the choice in a condition.

            djblets.conditions.errors.InvalidConditionValueError:
                The value was missing from the payload data or was not valid
                for the choice and operator in a condition.

            djblets.conditions.errors.InvalidConditionModeError:
                The stored match mode was missing or was not a valid mode.
        """
        mode = data.get('mode')

        if mode not in cls.CONDITIONS:
            logging.debug('ConditionSet.deserialize: Invalid "mode" value '
                          '"%s" for condition set %r',
                          mode, data)

            raise InvalidConditionModeError(
                _('"%s" is not a valid condition mode.')
                % mode)

        return cls(mode, [
            Condition.deserialize(choices, condition_data, i, choice_kwargs)
            for i, condition_data in enumerate(data.get('conditions', []))
        ])

    def __init__(self, mode=DEFAULT_MODE, conditions=[]):
        """Initialize the condition set.

        Args:
            mode (unicode, optional):
                The match mode. This defaults to :py:attr:`MODE_ALL`.

            conditions (list, optional):
                The conditions that make up this set. This defaults to an
                empty list.

        Raises:
            djblets.conditions.errors.InvalidConditionModeError:
                The match mode is not a valid mode.
        """
        if mode not in self.CONDITIONS:
            raise InvalidConditionModeError(
                _('"%s" is not a valid condition mode.')
                % mode)

        self.mode = mode
        self.conditions = conditions

    def matches(self, **values):
        """Check if a value matches the condition set.

        Depending on the mode of the condition set, this will either require
        all conditions to match, or only one.

        Args:
            **values (dict):
                Values to match against. By default, condition choices
                will match against a single ``value`` keyword argument, but
                more specialized uses might take into account one or more
                other keyword arguments.

        Returns:
            bool:
            ``True`` if the value fulfills the condition set. ``False`` if it
            does not.
        """
        if self.mode == self.MODE_ALWAYS:
            match_conditions = lambda x: True
        elif self.mode == self.MODE_ALL:
            match_conditions = self._match_all
        elif self.mode == self.MODE_ANY:
            match_conditions = any
        else:
            # We shouldn't be here, unless someone set the mode to a bad value
            # after creating the condition set.
            raise ValueError('Invalid condition mode %r' % self.mode)

        return match_conditions(self._get_condition_results(self.conditions,
                                                            values))

    def serialize(self):
        """Serialize the condition set to a JSON-serializable dictionary.

        Returns:
            dict:
            A dictionary representing the condition set. It can be safely
            serialized to JSON.
        """
        return {
            'mode': self.mode,
            'conditions': [
                condition.serialize()
                for condition in self.conditions
            ],
        }

    # Make this serializable in a DjbletsJSONEncoder.
    to_json = serialize

    def _get_condition_results(self, conditions, values):
        """Yield the results from each condition match.

        This will iterate through all the conditions, running a match against
        the provided values, yielding each result.

        If a condition expects a particular value that's not provided in
        ``values``, it will evaluate as a false match.

        Args:
            conditions (list of djblets.conditions.condition.Condition):
                The conditions to iterate through.

            values (dict):
                The dictionary of values to match against.

        Yields:
            bool:
            The result of each condition match.
        """
        value_state_cache = {}

        for condition in conditions:
            value_kwarg = condition.choice.value_kwarg

            if value_kwarg in values:
                yield condition.matches(values[value_kwarg],
                                        value_state_cache=value_state_cache)
            else:
                yield False

    def _match_all(self, results):
        """Return whether all results are truthy and the list is non-empty.

        This works similarly to :py:func:`all`, but will return ``False`` if
        the provided list is empty.

        Args:
            results (generator):
                The condition results to iterate through.

        Returns:
            bool:
            ``True`` if there are condition results present and they all
            evaluate to ``True``. ``False`` otherwise.
        """
        found = False

        for result in results:
            if not result:
                return False

            found = True

        return found



"""Unit tests for BaseWebAPIToken and WebAPITokenManager."""

import warnings
from datetime import timedelta

import kgb
from django.contrib.auth.models import User
from django.db import models
from django.utils import timezone

from djblets.deprecation import RemovedInDjblets40Warning
from djblets.secrets.token_generators.legacy_sha1 import \
    LegacySHA1TokenGenerator
from djblets.secrets.token_generators.vendor_checksum import \
    VendorChecksumTokenGenerator
from djblets.testing.testcases import TestCase
from djblets.webapi.errors import WebAPITokenGenerationError
from djblets.webapi.models import BaseWebAPIToken


class WebAPIToken(BaseWebAPIToken):
    my_field = models.BooleanField(default=False)


class WebAPITokenTests(TestCase):
    """Unit tests for BaseWebAPIToken.

    Version Added:
        3.0
    """

    def setUp(self):
        super(WebAPITokenTests, self).setUp()

        self.token_generator_id = \
            VendorChecksumTokenGenerator.token_generator_id
        self.token_info = {'token_type': 'test'}
        self.user = User.objects.create(username='test-user')

    def test_is_deprecated_with_deprecated(self):
        """Testing BaseWebAPIToken.is_deprecated with a token from a
        deprecated token generator
        """
        token_generator_id = LegacySHA1TokenGenerator.token_generator_id
        token_info = {
            'attempt': 0,
            'token_type': 'test',
            'user': self.user,
        }

        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            token_generator_id=token_generator_id,
            token_info=token_info)

        self.assertTrue(webapi_token.is_deprecated())

    def test_is_deprecated_with_non_deprecated(self):
        """Testing BaseWebAPIToken.is_deprecated with a token from a
        token generator that is not deprecated
        """
        token_generator_id = VendorChecksumTokenGenerator.token_generator_id
        token_info = {'token_type': 'test'}

        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            token_generator_id=token_generator_id,
            token_info=token_info)

        self.assertFalse(webapi_token.is_deprecated())

    def test_is_expired_with_expired(self):
        """Testing BaseWebAPIToken.is_expired with an expired token"""
        now = timezone.now()

        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            expires=(now - timedelta(hours=1)),
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        self.assertTrue(webapi_token.is_expired())

    def test_is_expired_with_non_expired(self):
        """Testing BaseWebAPIToken.is_expired with a non expired token"""
        now = timezone.now()

        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            expires=(now + timedelta(hours=1)),
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        self.assertFalse(webapi_token.is_expired())

    def test_is_expired_with_no_expiry(self):
        """Testing BaseWebAPIToken.is_expired with a token that has
        no expiration date set
        """
        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        self.assertFalse(webapi_token.is_expired())


class WebAPITokenManagerTests(kgb.SpyAgency, TestCase):
    """Unit tests for WebAPITokenManager."""

    def setUp(self):
        super(WebAPITokenManagerTests, self).setUp()

        self.token_generator_id = \
            VendorChecksumTokenGenerator.token_generator_id
        self.token_info = {'token_type': 'test'}
        self.user = User.objects.create(username='test-user')

    def test_generate_token_with_defaults(self):
        """Testing WebAPITokenManager.generate_token with default arguments"""
        self.spy_on(WebAPIToken.save, call_original=False)
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        webapi_token = WebAPIToken.objects.generate_token(self.user)

        with warnings.catch_warnings(record=True) as w:
            webapi_token = WebAPIToken.objects.generate_token(self.user)

            # Assert that there was a warning for each of the missing args.
            self.assertEqual(len(w), 2)
            self.assertEqual(w[0].category, RemovedInDjblets40Warning)
            self.assertEqual(str(w[0].message),
                             'The token_generator_id parameter was not set. '
                             'This parameter will be required in Djblets 4.0.')
            self.assertEqual(w[1].category, RemovedInDjblets40Warning)
            self.assertEqual(str(w[1].message),
                             'The token_info parameter was not set. '
                             'This parameter will be required in Djblets 4.0.')

        self.assertEqual(webapi_token.user, self.user)

        self.assertEqual(webapi_token.time_added, timezone.now())
        self.assertEqual(webapi_token.last_updated, timezone.now())
        self.assertIsNone(webapi_token.last_used)
        self.assertIsNone(webapi_token.expires)

        self.assertTrue(webapi_token.valid)
        self.assertIsNone(webapi_token.invalid_date)
        self.assertEqual(webapi_token.invalid_reason, '')

        self.assertEqual(webapi_token.policy, {})
        self.assertEqual(webapi_token.note, '')
        self.assertIsNotNone(webapi_token.token)
        self.assertEqual(webapi_token.token_generator_id,
                         LegacySHA1TokenGenerator.token_generator_id)
        self.assertFalse(webapi_token.my_field)

    def test_generate_token_with_custom_field(self):
        """Testing WebAPITokenManager.generate_token with custom field"""
        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            my_field=True,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        self.assertTrue(webapi_token.my_field)

    def test_last_updated(self):
        """Testing WebAPITokenManager.save updates the last updated field"""
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            last_updated=(timezone.now() - timedelta(hours=1)),
            time_added=(timezone.now() - timedelta(hours=1)),
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        webapi_token.save()

        self.assertEqual(webapi_token.last_updated, timezone.now())

    def test_generate_token_with_specific_generator(self):
        """Testing WebAPITokenManager.generate_token with specifying a
        token generator
        """
        webapi_token = WebAPIToken.objects.generate_token(
            self.user,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        self.assertEqual(webapi_token.token_generator_id,
                         self.token_generator_id)
        self.assertTrue(
            webapi_token.token.startswith(self.token_info['token_type']))

    def test_generate_token_with_nonexistant_generator(self):
        """Testing WebAPITokenManager.generate_token with specifying a
        token generator that does not exist in the registry
        """
        token_generator_id = 'missing_generator'
        token_info = {'token_type': 'test'}
        expected_message = ('Could not find the token generator '
                            'to generate the token.')
        expected_log = ('Could not find the missing_generator token '
                        'generator in the registry.')

        with self.assertRaisesMessage(WebAPITokenGenerationError,
                                      expected_message):
            with self.assertLogs() as logs:
                WebAPIToken.objects.generate_token(
                    self.user,
                    token_generator_id=token_generator_id,
                    token_info=token_info)

                self.assertEqual(logs.records[0].getMessage(),
                                 expected_log)

    def test_invalidate_token(self):
        """Testing WebAPITokenManager.invalidate_token"""
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        webapi_token1 = WebAPIToken.objects.generate_token(
            self.user,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        user2 = User.objects.create(username='test-user2')
        webapi_token2 = WebAPIToken.objects.generate_token(
            user2,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        WebAPIToken.objects.invalidate_token(token=webapi_token1.token)

        self._assert_token_is_invalid(webapi_token1)

        webapi_token2.refresh_from_db()
        self.assertTrue(webapi_token2.valid)

    def test_invalidate_token_with_reason(self):
        """Testing WebAPITokenManager.invalidate_token with an invalid
        reason given
        """
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        webapi_token1 = WebAPIToken.objects.generate_token(
            self.user,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        WebAPIToken.objects.invalidate_token(token=webapi_token1.token,
                                             invalid_reason='Revoked.')

        self._assert_token_is_invalid(webapi_token1, 'Revoked.')

    def test_invalidate_tokens(self):
        """Testing WebAPITokenManager.invalidate_tokens with a list of users"""
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        webapi_token1 = WebAPIToken.objects.generate_token(
            self.user,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        user2 = User.objects.create(username='test-user2')
        webapi_token2 = WebAPIToken.objects.generate_token(
            user2,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        user3 = User.objects.create(username='test-user3')
        webapi_token3 = WebAPIToken.objects.generate_token(
            user3,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        WebAPIToken.objects.invalidate_tokens(users=[self.user, user2],
                                              invalid_reason='Revoked.')

        self._assert_token_is_invalid(webapi_token1, 'Revoked.')
        self._assert_token_is_invalid(webapi_token2, 'Revoked.')

        webapi_token3.refresh_from_db()
        self.assertTrue(webapi_token3.valid)

    def test_invalidate_tokens_with_extra_query(self):
        """Testing WebAPITokenManager.invalidate_tokens with passing
        extra query parameters
        """
        self.spy_on(timezone.now, op=kgb.SpyOpReturn(timezone.now()))

        webapi_token1 = WebAPIToken.objects.generate_token(
            self.user,
            note='test',
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        user2 = User.objects.create(username='test-user2')
        webapi_token2 = WebAPIToken.objects.generate_token(
            user2,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        user3 = User.objects.create(username='test-user3')
        webapi_token3 = WebAPIToken.objects.generate_token(
            user3,
            token_generator_id=self.token_generator_id,
            token_info=self.token_info)

        extra_query = models.Q(note='test')

        WebAPIToken.objects.invalidate_tokens(users=[self.user, user2],
                                              extra_query=extra_query,
                                              invalid_reason='Revoked.')

        self._assert_token_is_invalid(webapi_token1, 'Revoked.')

        webapi_token2.refresh_from_db()
        self.assertTrue(webapi_token2.valid)

        webapi_token3.refresh_from_db()
        self.assertTrue(webapi_token3.valid)

    def _assert_token_is_invalid(self, token, invalid_reason=''):
        """Assert that the given token is invalid.

        Args:
            token (djblets.webapi.models.BaseWebAPIToken):
                The token.

            invalid_reason (str):
                The invalid reason that should be set on the token. This
                defaults to an empty string.

        Raises:
            AssertionError:
                The token is not invalid.
        """
        token.refresh_from_db()

        self.assertFalse(token.valid)
        self.assertEqual(token.invalid_date, timezone.now())
        self.assertEqual(token.invalid_reason, invalid_reason)



"""Unit tests for djblets.util.templatetags.djblets_js."""

from djblets.testing.testcases import TestCase
from djblets.util.templatetags.djblets_js import json_dumps


class JSONDumpsFilterTests(TestCase):
    """Unit tests for the {{...|json_dumps}} template filter."""

    def test_prevents_xss(self):
        """Testing {{...|json_dumps}} doesn't allow XSS injection"""
        # This is bug 3406.
        obj = {
            'xss': '</script><script>alert(1);</script>'
        }

        self.assertEqual(
            json_dumps(obj),
            '{"xss": "\\u003C/script\\u003E\\u003Cscript\\u003E'
            'alert(1);\\u003C/script\\u003E"}')

    def test_sorted_keys(self):
        """Testing {{...|json_dumps}} sorts dictionary keys"""
        self.assertEqual(
            json_dumps({
                'a': 1,
                'b': 2,
                'z': 3,
                '0': 4,
            }),
            '{"0": 4, "a": 1, "b": 2, "z": 3}')



from django.db.models.signals import post_save

from djblets.extensions.hooks import SignalHook
from djblets.integrations.integration import Integration
from djblets.integrations.manager import IntegrationManager
from djblets.integrations.tests.models import IntegrationConfig
from djblets.integrations.tests.testcases import IntegrationsTestCase


class DummyIntegration1(Integration):
    def initialize(self):
        self.dummy_initted = True
        self.hook = SignalHook(self, post_save, self._on_callback)

    def shutdown(self):
        self.dummy_initted = False

    def _on_callback(self, **kwargs):
        pass


class DummyIntegration2(Integration):
    def initialize(self):
        pass


class IntegrationTests(IntegrationsTestCase):
    """Unit tests for djblets.integrations.integration.Integration."""

    def setUp(self):
        super(IntegrationTests, self).setUp()

        self.manager = IntegrationManager(IntegrationConfig)
        self.integration = \
            self.manager.register_integration_class(DummyIntegration1)

    def test_init(self):
        """Testing Integration initialization"""
        self.assertEqual(self.integration.hooks, set())
        self.assertFalse(self.integration.enabled)
        self.assertFalse(hasattr(self.integration, 'dummy_initted'))

    def test_id(self):
        """Testing Integration.id"""
        self.assertEqual(self.integration.id, self.integration.integration_id)

    def test_enable_integration(self):
        """Testing Integration.enable_integration"""
        self.integration.enable_integration()

        self.assertTrue(self.integration.enabled)
        self.assertTrue(hasattr(self.integration, 'dummy_initted'))
        self.assertTrue(self.integration.dummy_initted)
        self.assertEqual(self.integration.hooks, {self.integration.hook})

    def test_disable_integration(self):
        """Testing Integration.disable_integration"""
        self.integration.enable_integration()
        self.integration.disable_integration()

        self.assertFalse(self.integration.enabled)
        self.assertTrue(hasattr(self.integration, 'dummy_initted'))
        self.assertFalse(self.integration.dummy_initted)
        self.assertEqual(self.integration.hooks, set())

    def test_get_configs(self):
        """Testing Integration.get_configs"""
        config1 = self.integration.create_config(name='Config 1',
                                                 enabled=True,
                                                 save=True)
        config2 = self.integration.create_config(name='Config 2',
                                                 enabled=True,
                                                 save=True)

        # Add some configs that shouldn't be returned.
        integration2 = \
            self.manager.register_integration_class(DummyIntegration2)
        self.integration.create_config(name='Config 3', save=True)
        integration2.create_config(name='Config 4', save=True)

        self.assertEqual(self.integration.get_configs(), [config1, config2])

    def test_get_configs_with_filter(self):
        """Testing Integration.get_configs with filter"""
        config1 = self.integration.create_config(name='Config 1',
                                                 enabled=True,
                                                 save=True)
        self.integration.create_config(name='Config 2',
                                       enabled=True,
                                       save=True)

        # Add some configs that shouldn't be returned.
        integration2 = \
            self.manager.register_integration_class(DummyIntegration2)
        self.integration.create_config(name='Config 3', save=True)
        integration2.create_config(name='Config 4', save=True)

        self.assertEqual(self.integration.get_configs(name='Config 1'),
                         [config1])

    def test_create_config(self):
        """Testing Integration.create_config"""
        config = self.integration.create_config(name='Config 1')
        self.assertFalse(config.enabled)
        self.assertIsNone(config.pk)

    def test_create_config_with_save(self):
        """Testing Integration.create_config with save=True"""
        config = self.integration.create_config(name='Config 1', save=True)
        self.assertFalse(config.enabled)
        self.assertIsNotNone(config.pk)



"""Support for tracking and looking up user consent for privacy rights.

This provides convenience imports for the following:

.. autosummary::
   :nosignatures:

   ~djblets.privacy.consent.base.BaseConsentRequirement
   ~djblets.privacy.consent.base.Consent
   ~djblets.privacy.consent.base.ConsentData
   ~djblets.privacy.consent.registry.get_consent_requirements_registry
   ~djblets.privacy.consent.tracker.BaseConsentTracker
   ~djblets.privacy.consent.tracker.DatabaseConsentTracker
   ~djblets.privacy.consent.tracker.get_consent_tracker
"""

from djblets.privacy.consent.base import (BaseConsentRequirement, Consent,
                                          ConsentData)
from djblets.privacy.consent.registry import get_consent_requirements_registry
from djblets.privacy.consent.tracker import (BaseConsentTracker,
                                             DatabaseConsentTracker,
                                             get_consent_tracker)


__all__ = (
    'BaseConsentRequirement',
    'BaseConsentTracker',
    'Consent',
    'ConsentData',
    'DatabaseConsentTracker',
    'get_consent_requirements_registry',
    'get_consent_tracker',
)

__autodoc_excludes__ = __all__
